from typing import List


class Detail:
    def __init__(self, count: int, sex: str, age: str) -> None:
        """Detail constructor
        :param count: Count
        :param sex: Sex
        :param age: Age
        :type count: int
        :type sex: str
        :type age: str
        """
        self.count: int = count
        self.sex: str = sex
        self.age: str = age

    @property
    def excel_str_german(self):
        excel_str = f"{self.count}x"
        if self.sex == "M":
            excel_str = f"{excel_str} Männchen"
        elif self.sex == "F":
            excel_str = f"{excel_str} Weibchen"
        elif self.sex == "FT":
            excel_str = (
                f"{excel_str} weibchenfarbig"
                if self.count == 1
                else f"{excel_str} weibchenfarbige"
            )

        if self.age == "PULL":
            excel_str = (
                f"{excel_str} Pullus / nicht-flügge"
                if self.count == 1
                else f"{excel_str} Pulli / nicht-flügge"
            )
        elif self.age == "1Y":
            excel_str = (
                f"{excel_str} 1. KJ / diesjährig"
                if self.count == 1
                else f"{excel_str} 1. KJ / diesjährige"
            )
        elif self.age == "2Y":
            excel_str = (
                f"{excel_str} 2. KJ / vorjährig"
                if self.count == 1
                else f"{excel_str} 2. KJ / vorjährige"
            )
        elif self.age == "3Y":
            excel_str = f"{excel_str} 3. KJ"
        elif self.age == "4Y":
            excel_str = f"{excel_str} 4. KJ"
        elif self.age == "5Y":
            excel_str = f"{excel_str} 5. KJ"
        elif self.age == "IMM":
            excel_str = (
                f"{excel_str} immatur" if self.count == 1 else f"{excel_str} immature"
            )
        elif self.age == "AD":
            excel_str = (
                f"{excel_str} adult" if self.count == 1 else f"{excel_str} adulte"
            )

        return excel_str

    def __str__(self) -> str:
        return f"{self.count}-{self.sex}-{self.age}"

    def __eq__(self, other):
        return (
            self.count == other.count
            and self.sex == other.sex
            and self.age == other.age
        )

    @staticmethod
    def list_to_excel_str(details: List["Detail"]):
        excel_str = ""
        for detail in details:
            if excel_str:
                excel_str = f"{excel_str} / {detail.excel_str_german}"
            else:
                excel_str = f"{detail.excel_str_german}"
        return excel_str



from typing import Optional

from ornitho.model.abstract import ListableModel
from ornitho.model.abstract.base_model import check_refresh
from ornitho.model.territorial_unit import TerritorialUnit


class LocalAdminUnit(ListableModel):
    ENDPOINT: str = "local_admin_units"

    def __init__(self, id_: int) -> None:
        super(LocalAdminUnit, self).__init__(id_)
        self._territorial_unit: Optional[TerritorialUnit] = None

    @property  # type: ignore
    @check_refresh
    def id_canton(self) -> int:
        return int(self._raw_data["id_canton"])

    @property  # type: ignore
    @check_refresh
    def name(self) -> str:
        return self._raw_data["name"]

    @property  # type: ignore
    @check_refresh
    def insee(self) -> str:
        return self._raw_data["insee"]

    @property  # type: ignore
    @check_refresh
    def coord_lon(self) -> float:
        return float(self._raw_data["coord_lon"])

    @property  # type: ignore
    @check_refresh
    def coord_lat(self) -> float:
        return float(self._raw_data["coord_lat"])

    @property
    def territorial_unit(self) -> TerritorialUnit:
        if self._territorial_unit is None:
            self._territorial_unit = TerritorialUnit(id_=self.id_canton)
        return self._territorial_unit

    @property
    def canton(self):
        return self.territorial_unit



# flake8: noqa
"""
    Ornitho
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    An ornitho API client
    :copyright: (c) 2019 by DDA
    :license: MIT, see LICENSE for more details.
"""
import logging
import os
from typing import Callable, Optional

import requests

from ornitho.api_exception import (
    APIConnectionException,
    APIException,
    APIHttpException,
    AuthenticationException,
    BadGatewayException,
    ContentTypeException,
    GatewayTimeoutException,
    ObjectNotFoundException,
    ServiceUnavailableException,
)
from ornitho.api_requester import APIRequester
from ornitho.model import (
    Detail,
    Entity,
    EstimationCode,
    Family,
    Field,
    FieldOption,
    Form,
    LocalAdminUnit,
    MapLayer,
    Media,
    ModificationType,
    Observation,
    Observer,
    Place,
    Precision,
    Protocol,
    Relation,
    RelationType,
    Right,
    Site,
    Source,
    Species,
    TaxonomicGroup,
    TerritorialUnit,
)

__version__ = "0.3.0"
__license__ = "MIT"


# Configuration variables

consumer_key: Optional[str] = None
consumer_secret: Optional[str] = None
user_email: Optional[str] = None
user_pw: Optional[str] = None
api_base: Optional[str] = None

app_info = None

cache_enabled: bool = False
cache_name: str = "ornitho_cache"
cache_backend: str = "sqlite"
cache_expire_after: int = 600
cache_filter_fn: Callable[[requests.Response], bool] = (
    lambda r: "pagination_key" not in r.headers.keys()
)
cache_redis_host: str = "localhost"
cache_redis_port: int = 6379
cache_redis_db: int = 0

log_level = os.environ.get("ORNITHO_LOG_LEVEL") or logging.WARNING
logging.basicConfig(
    level=log_level, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)



from typing import List, Optional, Union

from ornitho import APIException
from ornitho.api_requester import APIRequester
from ornitho.model.abstract import ListableModel
from ornitho.model.field_option import FieldOption


class Field(ListableModel):
    ENDPOINT: str = "fields"

    def __init__(self, id_: int) -> None:
        """Form constructor
        :param id_: ID, which is used to get the form from Biolovison
        :type id_: int
        """
        super(Field, self).__init__(id_)
        self._options: Optional[List[FieldOption]] = None

    @classmethod
    def get(cls, id_: Union[int, str], short_version: bool = False) -> "Field":
        """Retrieve Object from Biolovision with given ID
        :param id_: Unique identifier
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :type id_: Union[int, str]
        :type short_version: bool
        :return: Instance, retrieved from Biolovision with given ID
        :rtype: Field
        """
        fields = cls.list_all(short_version=short_version)
        for field in fields:
            if field.id_ == id_:
                return field
        raise APIException(f"Can't find field with ID {id_}")

    def refresh(
        self,
        short_version: bool = False,
        retries: int = 0,
    ) -> "Field":
        raise NotImplementedError

    @property
    def group(self) -> str:
        return self._raw_data["group"]

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def text(self) -> Optional[str]:
        return self._raw_data["text"] if "text" in self._raw_data else None

    @property
    def default(self) -> int:
        return int(self._raw_data["default"])

    @property
    def mandatory(self) -> bool:
        return False if self._raw_data.get("mandatory") == "0" else True

    @property
    def empty_choice(self) -> bool:
        return False if self._raw_data.get("empty_choice") == "0" else True

    @property
    def options(self) -> List[FieldOption]:
        if self._options is None:
            with APIRequester() as requester:
                url = f"fields/{self.id_}"
                response, pagination_key = requester.request(method="GET", url=url)
                self._options = [
                    FieldOption.create_from_ornitho_json(option) for option in response
                ]
        return self._options



# flake8: noqa

from ornitho.model.detail import Detail
from ornitho.model.entity import Entity
from ornitho.model.family import Family
from ornitho.model.field import Field
from ornitho.model.field_option import FieldOption
from ornitho.model.form import Form
from ornitho.model.local_admin_unit import LocalAdminUnit
from ornitho.model.media import Media
from ornitho.model.modification_type import ModificationType
from ornitho.model.observation import EstimationCode, Observation, Precision, Source
from ornitho.model.observer import Observer
from ornitho.model.place import Place
from ornitho.model.protocol import Protocol
from ornitho.model.relation import Relation, RelationType
from ornitho.model.right import Right
from ornitho.model.site import MapLayer, Site
from ornitho.model.species import Species
from ornitho.model.taxo_group import TaxonomicGroup
from ornitho.model.territorial_unit import TerritorialUnit



from abc import ABC
from typing import TypeVar

from ornitho.model.abstract import BaseModel

# Create a generic variable that can be 'CreateableModel', or any subclass.
T = TypeVar("T", bound="UpdateableModel")


class UpdateableModel(BaseModel, ABC):
    """Abstract class for updateable models via POST /ENDPOINT"""

    def update(
        self: T,
        retries: int = 0,
    ):
        """Update an instance on ornitho, respecting the specific model characteristics"""
        url = f"{self.ENDPOINT}/{self.id_}"
        body = {"data": {"sightings": [self.raw_data_trim_field_ids()]}}
        response = self.request(method="put", url=url, body=body, retries=retries)
        return response



from ornitho.model.abstract import ListableModel
from ornitho.model.abstract.base_model import check_refresh


class TerritorialUnit(ListableModel):
    ENDPOINT: str = "territorial_units"

    @property  # type: ignore
    @check_refresh
    def id_country(self) -> int:
        return int(self._raw_data["id_country"])

    @property  # type: ignore
    @check_refresh
    def name(self) -> str:
        return self._raw_data["name"]

    @property  # type: ignore
    @check_refresh
    def short_name(self) -> str:
        return self._raw_data["short_name"]



import json
from copy import deepcopy
from datetime import date, datetime
from json.decoder import JSONDecodeError
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode

from requests import Response
from requests_oauthlib import OAuth1Session

import ornitho
from ornitho import api_exception

try:
    import requests_cache

    class CachedOAuthSession(
        requests_cache.CacheMixin, OAuth1Session
    ):  # pragma: no cover
        """Session with features from both CachedSession and OAuth1Session"""

except ImportError:  # pragma: no cover
    requests_cache = None


class APIRequester(object):
    """Class for making API requests"""

    def __init__(
        self,
        consumer_key: Optional[str] = None,
        consumer_secret: Optional[str] = None,
        user_email: Optional[str] = None,
        user_pw: Optional[str] = None,
        api_base: Optional[str] = None,
    ) -> None:
        """API requester constructor
        :param consumer_key: Optional Consumer Key, overrides field from ornitho module (ornitho.consumer_key)
        :param consumer_secret: Optional Consumer Secret, overrides field from ornitho module (ornitho.consumer_secret)
        :param user_email: Optional User Mail, overrides field from ornitho module (ornitho.user_email)
        :param user_pw: Optional User Password, overrides field from ornitho module (ornitho.user_pw)
        :param api_base: Optional API base url, overrides field from ornitho module (ornitho.api_base)
        :type consumer_key: Optional[str]
        :type consumer_secret: Optional[str]
        :type user_email: Optional[str]
        :type user_pw: Optional[str]
        :type api_base: Optional[str]
        """
        self.consumer_key: Optional[str] = consumer_key or ornitho.consumer_key
        self.consumer_secret: Optional[str] = consumer_secret or ornitho.consumer_secret
        self.user_email: Optional[str] = user_email or ornitho.user_email
        self.user_pw: Optional[str] = user_pw or ornitho.user_pw
        self.api_base: Optional[str] = api_base or ornitho.api_base
        if not self.consumer_key:
            raise RuntimeError("consumer_key missing!")
        if not self.consumer_secret:
            raise RuntimeError("consumer_secret missing!")
        if not self.user_email:
            raise RuntimeError("user_email missing!")
        if not self.user_pw:
            raise RuntimeError("user_pw missing!")
        if not self.api_base:
            raise RuntimeError("api_base missing!")

        if ornitho.cache_enabled and requests_cache:  # pragma: no cover
            if ornitho.cache_backend == "redis":
                import redis

                r = redis.StrictRedis(
                    host=ornitho.cache_redis_host,
                    port=ornitho.cache_redis_port,
                    db=ornitho.cache_redis_db,
                )
                self.session: OAuth1Session = CachedOAuthSession(
                    client_key=self.consumer_key,
                    client_secret=self.consumer_secret,
                    cache_name=ornitho.cache_name,
                    backend=ornitho.cache_backend,
                    expire_after=ornitho.cache_expire_after,
                    filter_fn=ornitho.cache_filter_fn,
                    connection=r,
                )
            else:
                self.session = CachedOAuthSession(
                    client_key=self.consumer_key,
                    client_secret=self.consumer_secret,
                    cache_name=ornitho.cache_name,
                    backend=ornitho.cache_backend,
                    expire_after=ornitho.cache_expire_after,
                    filter_fn=ornitho.cache_filter_fn,
                )
        else:
            if ornitho.cache_enabled:  # pragma: no cover
                ornitho.logger.warning(
                    "Cache was enabled but caching dependency is not install. Please install 'requests_cache'"
                )
            self.session = OAuth1Session(
                client_key=self.consumer_key, client_secret=self.consumer_secret
            )

    def __enter__(self):
        """Used by a with-statement"""
        return self

    def __exit__(self, *args):
        """Used by a with-statement"""
        self.close()

    def close(self):
        """Close an OAuth1 Session"""
        self.session.close()

    def request(
        self,
        method: str,
        url: str,
        pagination_key: Optional[str] = None,
        short_version: bool = False,
        request_all: bool = False,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
        retries: int = 0,
    ) -> Tuple[Union[bytes, List[Dict[str, str]]], Optional[str]]:
        """Make requests to the API
        If request_all ist set, several requests calls to the API can be made, until all data is retrieved. Else a
        pagination key will be returned, which can be used to get the next data set.
        :param method: HTTP Method e.g. 'GET'
        :param url: API URL to call
        :param pagination_key: Additional pagination key, to get the next page
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
            Default: 'False'
        :param request_all:  Indicates, if all pages should be returned. May result in many API calls. Default: 'False'
        :param params: Additional URL parameters.
        :param body: Request body
        :param retries: Indicates how many retries should be performed
        :type method: str
        :type url: str
        :type pagination_key: str
        :type short_version: bool
        :type request_all: bool
        :type params: Dict[str, Any]
        :type body: Dict[str, Any]
        :type retries: int
        :return: Tuple of raw data list and pagination key
        :rtype: Tuple[List[Dict[str, str]], Optional[str]]
        """
        data: List[Dict[str, str]] = []
        responds, pk = self.request_raw(
            method=method.lower(),
            url=url,
            pagination_key=pagination_key,
            short_version=short_version,
            params=params,
            body=body,
            retries=retries,
        )
        if isinstance(responds, bytes):
            return responds, pk
        elif isinstance(responds, list):
            data += responds
        elif isinstance(responds, dict) and "data" in responds.keys():
            if isinstance(responds["data"], dict):
                if "sightings" in responds["data"]:
                    data += responds["data"]["sightings"]
                if "forms" in responds["data"]:
                    for form in responds["data"]["forms"]:
                        day = deepcopy(form["sightings"][0]["date"])
                        form_copy_wihtout_sightings = deepcopy(form)
                        form_copy_wihtout_sightings["day"] = day
                        del form_copy_wihtout_sightings["sightings"]
                        for sighting in form["sightings"]:
                            sighting["form"] = form_copy_wihtout_sightings
                            data.append(sighting)
            else:
                data += responds["data"]
        else:
            data.append(responds)

        ornitho.logger.info("Received %s data objects" % (len(data)))

        if pk and request_all and len(data) > 0:
            next_response = self.request(
                method=method,
                url=url,
                pagination_key=pk,
                short_version=short_version,
                request_all=request_all,
                params=params,
                body=body,
                retries=retries,
            )[0]
            if isinstance(next_response, bytes):
                raise api_exception.APIException(
                    "Received bytes content, where json was expected"
                )
            data = data + next_response
        if not pagination_key and len(data) <= 0:
            ornitho.logger.debug(
                "No data received! This can be caused by wrong parameter or an error in ornitho."
            )

        return data, pk

    @staticmethod
    def handle_error_response(response: Response) -> None:
        """Check the error response and raises a proper exception
        :param response: Erroneous response, received from the API
        :type response: Response
        :raise APIHttpException: Response contains unhandled HTTP Code
        :raise AuthenticationException: Authentication failed, wrong credentials?
        :raise BadGatewayException: Unknown Server error, in most cases a retry is successful
        :raise GatewayTimeoutException: Request took to long, reduce possible response by adding filters
        :raise ServiceUnavailableException: Service unavailable, in most cases a retry is successful
        """
        if response.status_code == 401:
            raise api_exception.AuthenticationException(response)
        elif response.status_code == 502:
            raise api_exception.BadGatewayException(response)
        elif response.status_code == 503:
            raise api_exception.ServiceUnavailableException(response)
        elif response.status_code == 504:
            raise api_exception.GatewayTimeoutException(response)
        else:
            raise api_exception.APIHttpException(response)

    @staticmethod
    def request_headers() -> Dict[str, str]:
        """Generate header information, like 'User-Agent'
        :return: Header information
        :rtype: Dict[str, str]
        """
        user_agent = f"API Python Client/{ornitho.__version__}"
        headers = {"User-Agent": user_agent}
        return headers

    def request_raw(
        self,
        method: str,
        url: str,
        pagination_key: Optional[str] = None,
        short_version: bool = False,
        params: Dict[str, Any] = None,
        body: Dict[str, Any] = None,
        retries: int = 0,
    ) -> Tuple[Any, Any]:
        """Make direct request to the API
        :param method: HTTP Method e.g. 'GET'
        :param url: API URL to call
        :param pagination_key: Additional pagination key, to get the next page
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
            Default: 'False'
        :param params: Additional URL parameters.
        :param body: Request body
        :param retries: Indicates how many retries should be performed
        :type method: str
        :type url: str
        :type pagination_key: str
        :type short_version: bool
        :type params: Dict[str, Any]
        :type body: Dict[str, Any]
        :type retries: int
        :return: Tuple of raw response data and pagination key
        :rtype: Tuple[Any, Any]
        :raise APIConnectionError: Unspecified error while connecting to Biolovision
        :raise APIException: Response contains unhandled HTTP Code
        :raise AuthenticationException: Authentication failed, wrong credentials?
        :raise GatewayTimeoutException: Request took to long, reduce possible response by adding filters
        :raise BadGatewayException: Unknown Server error, in most cases a retry is successful
        :raise ServiceUnavailableException: Service unavailable, in most cases a retry is successful
        :raise ContentTypeException: Unhandled Content Type received or no information about content typ found
        """

        params_dict: Dict[str, Any] = {
            "user_email": self.user_email,
            "user_pw": self.user_pw,
        }

        if pagination_key:
            params_dict["pagination_key"] = pagination_key

        if short_version:
            params_dict["short_version"] = 1

        if params:
            for key, value in params.items():
                if isinstance(value, bool):
                    value = 1 if value else 0
                    params[key] = value
                elif isinstance(value, datetime):
                    value = value.replace(microsecond=0)
                    if value.tzinfo:
                        value = value.astimezone(
                            datetime.now().astimezone().tzinfo
                        ).replace(tzinfo=None)
                    # ISO Format (especially time) is accepted but mostly ignored – only known exception is /observations/diff
                    # body[key] = value.replace(microsecond=0).isoformat()
                    # value = value.isoformat()
                    value = value.strftime("%d.%m.%Y")
                    params[key] = value
                elif isinstance(value, date):
                    value = value.strftime("%d.%m.%Y")
                    params[key] = value
                params_dict[key] = value

        abs_url = f"{self.api_base}{url}?{urlencode(params_dict)}"

        if body:
            for key, value in body.items():
                if isinstance(value, datetime):
                    # ISO Format (especially time) is accepted but mostly ignored
                    # body[key] = value.replace(microsecond=0).isoformat()
                    value = value.replace(microsecond=0)
                    if value.tzinfo:
                        value = value.astimezone(
                            datetime.now().astimezone().tzinfo
                        ).replace(tzinfo=None)
                    body[key] = value.strftime("%d.%m.%Y")
                elif isinstance(value, date):
                    body[key] = value.strftime("%d.%m.%Y")

        data = json.dumps(body) if body else None

        headers = self.request_headers()
        ornitho.logger.info(
            f"Request to Ornitho api. method={method}, url=/{url}, params={params}, short_version={short_version}, body={body}"
        )
        raw_response = self.session.request(method, abs_url, data=data, headers=headers)

        if not 200 <= raw_response.status_code < 300:
            if retries > 0:
                ornitho.logger.warning(
                    f"Response {raw_response.status_code}. {retries} left! Retry..."
                )
                return self.request_raw(
                    method=method,
                    url=url,
                    pagination_key=pagination_key,
                    short_version=short_version,
                    params=params,
                    body=body,
                    retries=retries - 1,
                )
            self.handle_error_response(raw_response)

        if (
            "pagination_key" in raw_response.headers.keys()
            and "Transfer-Encoding" in raw_response.headers.keys()
            and raw_response.headers["Transfer-Encoding"] == "chunked"
        ):
            pagination_key = raw_response.headers["pagination_key"] or None
        else:
            pagination_key = None

        if "Content-Type" in raw_response.headers.keys():
            if raw_response.headers["Content-Type"].startswith("application/json"):
                raw_response_text = raw_response.text
                # Remove the first JSON Line, which breaks the JSON format
                # Hopefully there aren't any more success messages
                # A real WTF moment...
                # TODO Add other language checks
                if (
                    raw_response_text.split("\n")[0]
                    == "API message : Ihre Beobachtungsdaten wurden erfolgreich übermittelt, vielen Dank!"
                ):
                    raw_response_text = "\n".join(raw_response_text.split("\n")[1:])
                try:
                    decoded_json_response = json.loads(raw_response_text)
                except JSONDecodeError:
                    raise api_exception.APIException(
                        f"Cant decode the response as JSON:\n{raw_response_text}"
                    )
                return decoded_json_response, pagination_key
            elif raw_response.headers["Content-Type"] == "application/pdf":
                return raw_response.content, pagination_key
            elif raw_response.headers["Content-Type"] == "text/html; charset=UTF-8":
                return raw_response.text, pagination_key
            else:
                raise api_exception.ContentTypeException(raw_response)
        else:
            raise api_exception.ContentTypeException(raw_response)



from typing import Optional

from ornitho.model.abstract import ListableModel
from ornitho.model.abstract.base_model import check_refresh
from ornitho.model.family import Family
from ornitho.model.taxo_group import TaxonomicGroup


class Species(ListableModel):

    ENDPOINT: str = "species"

    def __init__(self, id_: int) -> None:
        """Species constructor
        :param id_: ID, which is used to get the observation from Biolovison
        :type id_: int
        """
        super(Species, self).__init__(id_)
        self._taxo_group: Optional[TaxonomicGroup] = None
        self._family: Optional[Family] = None

    @property  # type: ignore
    @check_refresh
    def id_taxo_group(self) -> int:
        return int(self._raw_data["id_taxo_group"])

    @property  # type: ignore
    @check_refresh
    def sys_order(self) -> int:
        return int(self._raw_data["sys_order"])

    @property  # type: ignore
    @check_refresh
    def sempach_id_family(self) -> int:
        return int(self._raw_data["sempach_id_family"])

    @property  # type: ignore
    @check_refresh
    def category_1(self) -> str:
        return self._raw_data["category_1"]

    @property  # type: ignore
    @check_refresh
    def rarity(self) -> str:
        return self._raw_data["rarity"]

    @property  # type: ignore
    @check_refresh
    def atlas_start(self) -> int:
        return int(self._raw_data["atlas_start"])

    @property  # type: ignore
    @check_refresh
    def atlas_end(self) -> int:
        return int(self._raw_data["atlas_end"])

    @property  # type: ignore
    @check_refresh
    def latin_name(self) -> str:
        return self._raw_data["latin_name"]

    @property  # type: ignore
    @check_refresh
    def french_name(self) -> str:
        return self._raw_data["french_name"]

    @property  # type: ignore
    @check_refresh
    def french_name_plur(self) -> str:
        return self._raw_data["french_name_plur"]

    @property  # type: ignore
    @check_refresh
    def german_name(self) -> str:
        return self._raw_data["german_name"].replace("|", "")

    @property  # type: ignore
    @check_refresh
    def german_name_plur(self) -> str:
        return self._raw_data["german_name_plur"]

    @property  # type: ignore
    @check_refresh
    def english_name(self) -> str:
        return self._raw_data["english_name"]

    @property  # type: ignore
    @check_refresh
    def english_name_plur(self) -> str:
        return self._raw_data["english_name_plur"]

    @property  # type: ignore
    @check_refresh
    def is_used(self) -> bool:
        return False if self._raw_data.get("is_used") == "0" else True

    @property
    def taxo_group(self) -> TaxonomicGroup:
        if self._taxo_group is None:
            self._taxo_group = TaxonomicGroup.get(self.id_taxo_group)
        return self._taxo_group

    @property
    def family(self) -> Family:
        if self._family is None:
            self._family = Family.get(self.sempach_id_family)
        return self._family

    # Following Properties appear only when requesting an Observation. Mapping to the species API is done here
    @property
    def taxonomy(self) -> int:
        return (
            int(self._raw_data["taxonomy"])
            if "taxonomy" in self._raw_data
            else self.id_taxo_group
        )

    @property
    def category(self) -> str:
        return (
            self._raw_data["category"]
            if "category" in self._raw_data
            else self.category_1
        )

    @property
    def name(self) -> str:
        return self._raw_data["name"] if "name" in self._raw_data else self.german_name

    @property
    def dda_id_species(self) -> Optional[int]:
        return (
            int(self._raw_data["dda_id_species"]["#text"])
            if "dda_id_species" in self._raw_data
            else None
        )

    @property
    def euring_id_species(self) -> Optional[int]:
        return (
            int(self._raw_data["euring_id_species"]["#text"])
            if "euring_id_species" in self._raw_data
            else None
        )



from abc import ABC
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from ornitho.api_exception import ObjectNotFoundException
from ornitho.api_requester import APIRequester

# Create a generic variable that can be 'BaseModel', or any subclass.
T = TypeVar("T", bound="BaseModel")


class BaseModel(ABC):
    """Abstract base class for all models"""

    ENDPOINT: str

    def __init__(self, id_: Union[int, str] = None):
        """Base model constructor
        :param id_: Unique identifier
        :type id_: Union[int, str]
        """
        super(BaseModel, self).__init__()
        self._id: Optional[Union[int, str]] = id_
        self._raw_data: Dict[str, Any] = dict()
        self._previous: Dict[str, Any] = dict()
        self._refreshed: bool = False

    @property
    def id_(self) -> Optional[Union[int, str]]:
        """Unique identifier"""
        if self._id is None:
            if "id" in self._raw_data:
                self._id = self._raw_data["id"]
            elif "@id" in self._raw_data:
                self._id = self._raw_data["@id"]
        return self._id

    @classmethod
    def get(
        cls: Type[T],
        id_: Union[int, str],
        short_version: bool = False,
    ) -> T:
        """Retrieve Object from Biolovision with given ID
        :param id_: Unique identifier
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :type id_: Union[int, str]
        :type short_version: bool
        :return: Instance, retrieved from Biolovision with given ID
        :rtype: T
        """
        instance = cls(id_)
        instance.refresh(short_version=short_version)
        return instance

    @staticmethod
    def request(
        method: str,
        url: str,
        params: Dict[str, Any] = None,
        body: Dict[str, Any] = None,
        short_version: bool = False,
        retries: int = 0,
    ) -> List[Any]:
        """Send request to Biolovision and returns response
        :param method: HTTP Method (e.g. 'GET', 'POST', ...)
        :param url: Url to request
        :param params: Additional URL parameters.
        :param body: Request body
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :type method: str
        :type url: str
        :type params: Dict[str, Any]
        :type body: Dict[str, Any]
        :type short_version: bool
        :type retries: int
        :return: Response map from Biolovision
        :rtype: List[Any]
        """
        with APIRequester() as requester:
            response, pagination_key = requester.request(
                method=method,
                url=url,
                params=params,
                body=body,
                short_version=short_version,
                retries=retries,
            )
        # noinspection PyTypeChecker
        return response

    @classmethod
    def create_from_ornitho_json(cls: Type[T], data: Dict[str, Any]) -> T:
        identifier: Union[int, str]
        if "@id" in data:
            identifier = int(data["@id"]) if data["@id"].isdigit() else data["@id"]
        else:
            identifier = int(data["id"]) if data["id"].isdigit() else data["id"]
        obj = cls(identifier)
        obj._raw_data = data
        return obj

    def refresh(self: T, short_version: bool = False, retries: int = 0) -> T:
        """Refresh local model
        Call the api and refresh fields from response
        :return: Refreshed Object
        :rtype: T
        :raise ObjectNotFoundException: No or more than one objects retrieved
        """
        data = self.request(
            method="GET",
            url=self.instance_url(),
            short_version=short_version,
            retries=retries,
        )
        if len(data) != 1:
            raise ObjectNotFoundException(
                f"Get {len(data)} objects for {self.instance_url()}"
            )
        self._previous = self._raw_data
        self._raw_data = data[0]
        self._refreshed = True
        return self

    def instance_url(self) -> str:
        """Returns url for this instance
        :return: Instance's url
        :rtype: str
        """
        return f"{self.ENDPOINT}/{self.id_}"

    def raw_data_trim_field_ids(self) -> Dict[str, Any]:
        """Returns raw data with removed field id from field options ('2_3' -> '3').
        Only need to be reimplemented by models that use fields and are createable/updateable.
        :return: Raw data dict, with trimmed field options
        :rtype: Dict[str, Any]
        """
        return self._raw_data


def check_refresh(func):
    def wrapper(self: T):
        if func.__name__ not in self._raw_data and not self._refreshed:
            self.refresh()
        return func(self)

    return wrapper


def check_raw_data(key):
    def decorator(func):
        def wrapper(self: T):
            if key not in self._raw_data and not self._refreshed:
                if self._id is not None:
                    self.refresh()
            return func(self)

        return wrapper

    return decorator



from datetime import datetime
from typing import Optional

from ornitho.model.abstract import BaseModel


class Media(BaseModel):
    ENDPOINT: str = "media"

    @property
    def obid(self) -> int:
        return int(self._raw_data["obid"])

    @property
    def obs_hidden(self) -> bool:
        return False if self._raw_data["obs_hidden"] == "0" else True

    @property
    def surname(self) -> str:
        return self._raw_data["surname"]

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def advanced_observer(self) -> bool:
        return False if self._raw_data["advanced_observer"] == "0" else True

    @property
    def traid(self) -> int:
        return int(self._raw_data["traid"])

    @property
    def tra_hidden(self) -> bool:
        return False if self._raw_data["tra_hidden"] == "0" else True

    @property
    def tra_surname(self) -> str:
        return self._raw_data["tra_surname"]

    @property
    def tra_name(self) -> str:
        return self._raw_data["tra_name"]

    @property
    def obs_power_user(self) -> bool:
        return False if self._raw_data["obs_power_user"] == "0" else True

    @property
    def tra_power_user(self) -> bool:
        return False if self._raw_data["tra_power_user"] == "0" else True

    @property
    def media(self) -> str:
        return self._raw_data["media"]

    @property
    def has_large(self) -> bool:
        return False if self._raw_data["has_large"] == "0" else True

    @property
    def insert_date(self) -> datetime:
        timing = datetime.fromtimestamp(
            int(self._raw_data["insert_date"]["@timestamp"]),
        ).astimezone()
        return timing

    @property
    def photo(self) -> str:
        if self.has_large and self.media == "PHOTO":
            return self._raw_data["photo"].replace("xsmall", "large")
        elif self.media == "PHOTO":
            return self._raw_data["photo"].replace("xsmall/", "")
        else:
            return self._raw_data["photo"]

    @property
    def photo_small(self) -> Optional[str]:
        return self._raw_data["photo"] if self.media == "PHOTO" else None



from ornitho.model.abstract import ListableModel


class TaxonomicGroup(ListableModel):
    ENDPOINT: str = "taxo_groups"

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def latin_name(self) -> str:
        return self._raw_data["latin_name"]

    @property
    def name_constant(self) -> str:
        return self._raw_data["name_constant"]

    @property
    def access_mode(self) -> str:
        return self._raw_data["access_mode"]



from enum import Enum


class RelationType(Enum):
    SAME = "same"
    PROBABLE = "probable"
    DIFF = "diff"


class Relation:
    def __init__(self, with_id: int, type: RelationType) -> None:
        """Detail constructor
        :param with_id: Id of the related observation
        :param type: Type of the relation
        :type with_id: int
        :type type: RelationType
        """
        self.with_id: int = with_id
        self.type: RelationType = type

    def __str__(self) -> str:
        return f"{self.with_id}-{self.type.value}"

    def __eq__(self, other):
        return self.with_id == other.with_id and self.type == other.type



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Type, TypeVar

from ornitho.model.abstract import BaseModel

# Create a generic variable that can be 'CreateableModel', or any subclass.
T = TypeVar("T", bound="CreateableModel")


class CreateableModel(BaseModel, ABC):
    """Abstract class for createable models via POST /ENDPOINT"""

    CREATE_ENDPOINT: Optional[str] = None

    @classmethod
    def create_in_ornitho(cls: Type[T], data: Dict[str, Any], retries: int = 0) -> int:
        """Create an instance on ornitho
        :param data: Data
        :param retries: Indicates how many retries should be performed
        :type data: T
        :type retries: int
        :return:  Ornitho ID of the created object
        :rtype: T
        """
        url = cls.CREATE_ENDPOINT if cls.CREATE_ENDPOINT is not None else cls.ENDPOINT
        body = {"data": data}
        response = cls.request(method="post", url=url, body=body, retries=retries)
        return response[0]["id"][0]

    @classmethod
    @abstractmethod
    def create(cls: Type[T], **kwargs) -> T:
        """Create an instance on ornitho, respecting the specific model characteristics

        This method should implement an easy to use way for creating new instances on ornitho,
        respecting the specific model characteristics. I should use the general create_in_ornitho method.
        """



from typing import List, Union

from ornitho import APIRequester


class Right:
    ENDPOINT: str = "observers/rights"

    def __init__(self, id_: int, name: str, comment: str) -> None:
        """Detail constructor
        :param id: ID
        :param name: Name
        :param comment: Comment
        :type id: int
        :type name: str
        :type comment: str
        """
        self.id_: int = id_
        self.name: str = name
        self.comment: str = comment

    def __str__(self) -> str:
        return f"{self.id_}-{self.name}-{self.comment}"

    @classmethod
    def retrieve_for_observer(cls, id_observer: Union[int, str]) -> List["Right"]:
        with APIRequester() as requester:
            url = f"{cls.ENDPOINT}/{id_observer}"
            response, pk = requester.request_raw(
                method="get",
                url=url,
            )
        return [
            cls(id_=int(right["id"]), name=right["name"], comment=right["comment"])
            for right in response["data"]["rights"]
        ]



from enum import Enum
from typing import List, Optional

from ornitho.api_requester import APIRequester
from ornitho.model.abstract.base_model import BaseModel, check_raw_data, check_refresh
from ornitho.model.observer import Observer
from ornitho.model.place import Place


class MapLayer(Enum):
    BKG = "BKG"
    TOPO_PLUS_OPEN = "TOPO_PLUS_OPEN"
    OSM2014 = "OSM2014"
    OSMLIVE = "OSMLIVE"


class Site(BaseModel):
    ENDPOINT: str = "protocol/sites"

    def __init__(self, id_: int) -> None:
        """Site constructor
        :param id_: ID, which is used to get the site from Biolovison
        :type id_: int
        """
        super(Site, self).__init__(id_)
        self._place: Optional[Place] = None
        self._point_places: Optional[List[Place]] = None
        self._transect_places: Optional[List[Place]] = None
        self._polygon_places: Optional[List[Place]] = None
        self._pdf: Optional[bytes] = None

    @property  # type: ignore
    @check_refresh
    def id_universal(self) -> str:
        return self._raw_data["id_universal"]

    @property  # type: ignore
    @check_refresh
    def custom_name(self) -> str:
        return self._raw_data["custom_name"]

    @property  # type: ignore
    @check_refresh
    def local_name(self) -> Optional[str]:
        return self._raw_data["local_name"] if "local_name" in self._raw_data else None

    @property  # type: ignore
    @check_refresh
    def id_reference_locality(self) -> int:
        return int(self._raw_data["id_reference_locality"])

    @property  # type: ignore
    @check_refresh
    def reference_locality(self) -> str:
        return self._raw_data["reference_locality"]

    @property  # type: ignore
    @check_refresh
    def id_protocol(self) -> int:
        return int(self._raw_data["id_protocol"])

    @property
    def place(self) -> Place:
        if self._place is None:
            self._place = Place(id_=self.id_reference_locality)
        return self._place

    @property  # type: ignore
    @check_raw_data("transects")
    def transect_places(self) -> Optional[List[Place]]:
        if self._transect_places is None:
            if "transects" in self._raw_data:
                self._transect_places = [
                    Place.create_from_site(raw_transect)
                    for raw_transect in self._raw_data["transects"]
                ]
        return self._transect_places

    @property  # type: ignore
    @check_raw_data("points")
    def point_places(self) -> Optional[List[Place]]:
        if self._point_places is None:
            if "points" in self._raw_data:
                self._point_places = [
                    Place.create_from_site(raw_transect)
                    for raw_transect in self._raw_data["points"]
                ]
        return self._point_places

    @property  # type: ignore
    @check_raw_data("polygons")
    def polygon_places(self) -> Optional[List[Place]]:
        if self._polygon_places is None:
            if "polygons" in self._raw_data:
                self._polygon_places = [
                    Place.create_from_site(raw_transect)
                    for raw_transect in self._raw_data["polygons"]
                ]
        return self._polygon_places

    @property  # type: ignore
    @check_raw_data("boundary_wkt")
    def boundary_wkt(self) -> Optional[str]:
        return (
            self._raw_data["boundary_wkt"] if "boundary_wkt" in self._raw_data else None
        )

    @property  # type: ignore
    @check_refresh
    def observers(self) -> List[Observer]:
        observers = []
        if "observers" in self._raw_data:
            observers = [
                Observer(id_=int(observer_id))
                for observer_id in self._raw_data["observers"]
            ]
        return observers

    def pdf(
        self,
        map_layer: MapLayer = None,
        greyscale: bool = False,
        greyline: bool = False,
        alpha: bool = False,
        boundary: bool = False,
        retries: int = 0,
    ) -> bytes:
        """Request the site pdf map
        :param map_layer: Used map layer
        :param greyscale: Switch for a greyscaled map layer
        :param greyline: Switch for a grey dashed route
        :param alpha: Switch for 50% transparency
        :param boundary: Switch for disable the outer buffer
        :param retries: Indicates how many retries should be performed
        :type map_layer: MapLayer
        :type greyscale: bool
        :type greyline: bool
        :type alpha: bool
        :type boundary: bool
        :type retries: int
        :return: PDF as bytes
        :rtype: bytes
        """
        with APIRequester() as requester:
            url = f"{self.instance_url()}.pdf"
            params = {"id": self.id_}
            if map_layer:
                params["map_layer"] = map_layer.value
            if greyscale:
                params["greyscale"] = 1
            if greyline:
                params["greyline"] = 1
            if alpha:
                params["alpha"] = 1
            if boundary:
                params["boundary"] = 1
            response, pagination_key = requester.request(
                method="GET", url=url, params=params, retries=retries
            )
        return response



from urllib.parse import parse_qs, urlsplit

from requests import Response


class APIException(Exception):
    """Base API exception class"""


class APIHttpException(APIException):
    """HTTP exception class"""

    def __init__(self, response: Response) -> None:
        """API Exception constructor"""
        super().__init__(response)
        self.response: Response = response

    def __str__(self) -> str:
        """Readable string representation"""
        return (
            f"{self.path} with query {self.query} and body {self.request_body} responses with {self.http_status}: "
            f"{self.reason}: {self.body}"
        )

    def __repr__(self) -> str:
        """Unambiguous string representation"""
        return (
            f"{self.__class__.__name__}(http_status={self.http_status}, reason='{self.reason}', "
            f"body='{self.body}', path='{self.path}', query='{self.query}', request_body='{self.request_body}')"
        )

    @property
    def http_status(self) -> int:
        """HTTP return code"""
        return self.response.status_code

    @property
    def reason(self) -> str:
        """Textual HTTP code"""
        return self.response.reason

    @property
    def body(self) -> str:
        """Response body as string"""
        return self.response.text

    @property
    def path(self) -> str:
        """Path of the failed request"""
        return str(urlsplit(self.response.request.url).path)

    @property
    def query(self) -> dict:
        """Path of the failed request"""
        query = parse_qs(str(urlsplit(self.response.request.url).query))
        # Remove sensible data
        if "user_email" in query.keys():
            query["user_email"] = ["***"]
        if "user_pw" in query.keys():
            query["user_pw"] = ["***"]
        return query

    @property
    def request_body(self) -> str:
        """Path of the failed request"""
        return str(self.response.request.body)


class APIConnectionException(APIHttpException):
    """Exception class indicating an connection error"""


class AuthenticationException(APIHttpException):
    """Exception class indicating failed authentication"""


class GatewayTimeoutException(APIHttpException):
    """Exception class indicating a gateway timeout error"""


class BadGatewayException(APIHttpException):
    """Exception class indicating a bad gateway error"""


class ServiceUnavailableException(APIHttpException):
    """Exception class indicating that api is unavailable"""


class ContentTypeException(APIHttpException):
    """Content Type exception"""

    def __str__(self) -> str:
        """Readable string representation"""
        return f"Unhandled Content-Typ '{self.content_type}' received! Received body: {self.body}"

    def __repr__(self) -> str:
        """Unambiguous string representation"""
        return f"{super().__repr__()}, content_type={self.content_type}"

    @property
    def content_type(self) -> str:
        """Path of the failed request"""
        return self.response.headers["Content-Type"]


class ObjectNotFoundException(APIException):
    """Object not found exception class"""



import re
from typing import Union

from ornitho import APIException
from ornitho.api_requester import APIRequester
from ornitho.model.abstract import ListableModel


class FieldOption(ListableModel):
    ENDPOINT: str = ""

    @classmethod
    def get(cls, id_: Union[int, str], short_version: bool = False) -> "FieldOption":
        """Retrieve Object from Biolovision with given ID
        :param id_: Unique identifier
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :type id_: str
        :type short_version: bool
        :return: Instance, retrieved from Biolovision with given ID
        :rtype: FieldOption
        """
        with APIRequester() as requester:
            if isinstance(id_, str) and re.match(r"(.*)_(.*)", id_):
                url = f"fields/{id_.split('_')[0]}"
            else:
                raise APIException("ID must be string matching (.*)_(.*)")
            response, pagination_key = requester.request(
                method="GET", url=url, short_version=short_version
            )
            for option in response:
                field_option = cls.create_from_ornitho_json(option)
                if field_option.id_ == id_:
                    return field_option
            raise APIException(f"Can't find field option with ID {id_}")

    def refresh(self):
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def text(self) -> str:
        return self._raw_data["text"]

    @property
    def value(self) -> int:
        return int(self._raw_data["value"])

    @property
    def order_id(self) -> int:
        return int(self._raw_data["order_id"])



from abc import ABC
from datetime import date, datetime
from typing import List, Optional, Tuple, Type, TypeVar, Union

from ornitho.api_requester import APIRequester
from ornitho.model.abstract import BaseModel

# Create a generic variable that can be 'SearchableModel', or any subclass.
T = TypeVar("T", bound="SearchableModel")


class SearchableModel(BaseModel, ABC):
    """Abstract class for searchable models via POST /ENDPOINT/search"""

    @classmethod
    def search(
        cls: Type[T],
        request_all: Optional[bool] = False,
        pagination_key: Optional[str] = None,
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool, date, datetime, list]
    ) -> Tuple[List[T], Optional[str]]:
        """Search for instances at Biolovision via POST search
        If the list is chunked, a pagination key ist returned
        :param request_all: Indicates, if all instances should be retrieved (may result in many API calls)
        :param pagination_key: Pagination key, which can be used to retrieve the next page
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :type short_version: bool
        :type retries: int
        :param kwargs: Search values
        :type kwargs: Union[str, int, float, bool]
        :type body: Dict[str, Any]s
        :return: Tuple of instances and pagination key
        :rtype: Tuple[List[T], Optional[str]]
        """
        with APIRequester() as requester:
            url = "%s/search" % cls.ENDPOINT
            response, pk = requester.request(
                method="post",
                url=url,
                request_all=request_all,
                pagination_key=pagination_key,
                short_version=short_version,
                body=kwargs,
                retries=retries,
            )
            model_list: List[T] = []
            for ele in response:
                obj = cls.create_from_ornitho_json(ele)
                model_list.append(obj)
        return model_list, pk

    @classmethod
    def search_all(
        cls: Type[T],
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool, date, datetime, list]
    ) -> List[T]:
        """Search for instances at Biolovision via POST search
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :param kwargs: Search values
        :type short_version: bool
        :type retries: int
        :type kwargs: Union[str, int, float, bool, list]
        :return: List of instances
        :rtype: List[T]
        """
        object_list, pk = cls.search(
            request_all=True,
            pagination_key=None,
            short_version=short_version,
            retries=retries,
            **kwargs
        )
        return object_list



from datetime import date, datetime
from typing import List, Optional, Tuple, Union

from ornitho.model.abstract import ListableModel
from ornitho.model.entity import Entity
from ornitho.model.observation import Observation
from ornitho.model.site import Site


class Protocol(ListableModel):
    ENDPOINT: str = "protocol"

    def __init__(self, id_: int) -> None:
        """Protocol constructor
        :param id_: ID, which is used to get the protocol from Biolovison
        :type id_: int
        """
        super(Protocol, self).__init__(id_)
        self._entity: Optional[Entity] = None
        self._sites: Optional[List[Site]] = None

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def nbre_points_min(self) -> int:
        return int(self._raw_data["nbre_points_min"])

    @property
    def nbre_points_max(self) -> int:
        return int(self._raw_data["nbre_points_max"])

    @property
    def time_point(self) -> Optional[int]:
        return (
            int(self._raw_data["time_point"]) if self._raw_data["time_point"] else None
        )

    @property
    def taxo_point(self) -> int:
        return int(self._raw_data["taxo_point"])

    @property
    def additional_taxo_point(self) -> Optional[int]:
        return (
            int(self._raw_data["additional_taxo_point"])
            if self._raw_data["additional_taxo_point"]
            else None
        )

    @property
    def nbre_transects_min(self) -> int:
        return int(self._raw_data["nbre_transects_min"])

    @property
    def nbre_transects_max(self) -> int:
        return int(self._raw_data["nbre_transects_max"])

    @property
    def time_transect(self) -> Optional[int]:
        return (
            int(self._raw_data["time_transect"])
            if self._raw_data["time_transect"]
            else None
        )

    @property
    def taxo_transect(self) -> int:
        return int(self._raw_data["taxo_transect"])

    @property
    def additional_taxo_transect(self) -> Optional[int]:
        return (
            int(self._raw_data["additional_taxo_transect"])
            if self._raw_data["additional_taxo_transect"]
            else None
        )

    @property
    def nbre_polygones_min(self) -> int:
        return int(self._raw_data["nbre_polygones_min"])

    @property
    def nbre_polygones_max(self) -> int:
        return int(self._raw_data["nbre_polygones_max"])

    @property
    def time_polygone(self) -> Optional[int]:
        return (
            int(self._raw_data["time_polygone"])
            if self._raw_data["time_polygone"]
            else None
        )

    @property
    def taxo_poly(self) -> int:
        return int(self._raw_data["taxo_poly"])

    @property
    def additional_taxo_poly(self) -> Optional[int]:
        return (
            int(self._raw_data["additional_taxo_poly"])
            if self._raw_data["additional_taxo_poly"]
            else None
        )

    @property
    def project_id(self) -> Optional[int]:
        return (
            int(self._raw_data["project_id"]) if self._raw_data["project_id"] else None
        )

    @property
    def id_entity(self) -> int:
        return int(self._raw_data["id_entity"])

    @property
    def nbre_bounding_box_max(self) -> int:
        return int(self._raw_data["nbre_bounding_box_max"])

    @property
    def nbre_passage(self) -> Optional[int]:
        return (
            int(self._raw_data["nbre_passage"])
            if self._raw_data["nbre_passage"]
            else None
        )

    @property
    def auto_hidden(self) -> int:
        return False if self._raw_data["auto_hidden"] == "0" else True

    @property
    def only_admin_create(self) -> int:
        return False if self._raw_data["only_admin_create"] == "0" else True

    @property
    def start_month(self) -> int:
        return int(self._raw_data["start_month"])

    @property
    def default_atlas_code(self) -> Optional[int]:
        return (
            int(self._raw_data["default_atlas_code"])
            if self._raw_data["default_atlas_code"]
            else None
        )

    @property
    def default_count(self) -> Optional[int]:
        return (
            int(self._raw_data["default_count"])
            if self._raw_data["default_count"]
            else None
        )

    @property
    def entity(self) -> Entity:
        """Entity of the protocol"""
        if self._entity is None:
            self._entity = Entity.get(self.id_entity)
        return self._entity

    @property
    def sites(self) -> List[Site]:
        """Get sites linked to the protocol
        :return: List of sites
        :rtype: List[Site]
        """
        if not self._sites:
            url = f"{self.ENDPOINT}/sites"
            params = {"id_protocol": self.id_}

            sites_object = self.request(method="get", url=url, params=params)[0]
            self._sites = [Site(id_=site_id) for site_id in sites_object.keys()]
        return self._sites

    def get_observations(
        self,
        request_all: Optional[bool] = False,
        pagination_key: Optional[str] = None,
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool, date, datetime],
    ) -> Tuple[List[Observation], Optional[str]]:
        """Get observations linked to the protocol
        The same search parameters can be used as for the observations (except 'only_protocol' which is automatically set)
        If the list is chunked, a pagination key ist returned
        :param request_all: Indicates, if all instances should be retrieved (may result in many API calls)
        :param pagination_key: Pagination key, which can be used to retrieve the next page
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :param kwargs: Additional filter values
        :type request_all: Optional[bool]
        :type pagination_key: Optional[str]
        :type short_version: bool
        :type retries: int
        :type kwargs: Union[str, int, float, bool, datetime]
        :return: List of observations
        :rtype: Tuple[List[Observation], Optional[str]]
        """
        if "period_choice" not in kwargs.keys():
            kwargs["period_choice"] = "all"
        return Observation.search(
            request_all=request_all,
            pagination_key=pagination_key,
            short_version=short_version,
            only_protocol=self.name,
            retries=retries,
            **kwargs,
        )

    def get_all_observations(
        self,
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool, date, datetime],
    ) -> List[Observation]:
        """Get observations linked to the protocol
        The same search parameters can be used as for the observations (except 'only_protocol' which is automatically set)
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :param kwargs: Additional filter values
        :type short_version: bool
        :type retries: int
        :type kwargs: Union[str, int, float, bool, datetime]
        :return: List of observations
        :rtype: List[Observation]
        """
        return self.get_observations(
            request_all=True,
            pagination_key=None,
            short_version=short_version,
            retries=retries,
            **kwargs,
        )[0]



from ornitho.model.abstract import ListableModel


class Entity(ListableModel):
    ENDPOINT: str = "entities"

    @property
    def short_name(self) -> str:
        return self._raw_data["short_name"]

    @property
    def full_name_german(self) -> str:
        return self._raw_data["full_name_german"]

    @property
    def address(self) -> str:
        return self._raw_data["address"]

    @property
    def url(self) -> str:
        return self._raw_data["url"]

    @property
    def description_german(self) -> str:
        return self._raw_data["description_german"]



import uuid
from copy import deepcopy
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import ornitho
import ornitho.model.form
from ornitho import APIException
from ornitho.model.abstract import (
    BaseModel,
    CreateableModel,
    DeletableModel,
    ListableModel,
    SearchableModel,
    UpdateableModel,
)
from ornitho.model.abstract.base_model import check_raw_data
from ornitho.model.detail import Detail
from ornitho.model.field_option import FieldOption
from ornitho.model.media import Media
from ornitho.model.modification_type import ModificationType
from ornitho.model.observer import Observer
from ornitho.model.place import Place
from ornitho.model.relation import Relation, RelationType
from ornitho.model.species import Species


class EstimationCode(Enum):
    EXACT_VALUE = "EXACT_VALUE"
    ESTIMATION = "ESTIMATION"
    MINIMUM = "MINIMUM"
    NO_VALUE = "NO_VALUE"


class Precision(Enum):
    PRECISE = "precise"
    SQUARE = "square"
    PLACE = "place"
    MUNICIPALITY = "municipality"
    POLYGON = "polygone"
    TRANSECT = "transect"
    SUBPLACE = "subplace"
    POLYGON_PRECISE = "polygone_precise"
    TRANSECT_PRECISE = "transect_precise"
    SUBPLACE_PRECISE = "subplace_precise"


class Source(Enum):
    WEB = "WEB"
    MOBILE_DELAYED = "MOBILE_DELAYED"
    MOBILE_LIVE = "MOBILE_LIVE"
    MOBILE_FORM_LIVE = "MOBILE_FORM_LIVE"
    MOBILE_FORM_DELAYED = "MOBILE_FORM_DELAYED"
    MOBILE_LIVE_IOS = "MOBILE_LIVE_IOS"
    MOBILE_DELAYED_IOS = "MOBILE_DELAYED_IOS"
    MOBILE_FORM_LIVE_IOS = "MOBILE_FORM_LIVE_IOS"
    MOBILE_FORM_DELAYED_IOS = "MOBILE_FORM_DELAYED_IOS"


class Observation(
    ListableModel, SearchableModel, CreateableModel, DeletableModel, UpdateableModel
):
    """Representation of on Observation"""

    ENDPOINT: str = "observations"

    def __init__(
        self, id_: int = None, modification_type: ModificationType = None
    ) -> None:
        """Observation constructor
        :param id_: ID, which is used to get the observation from Biolovison – None if a new observation will be created
        :param modification_type: Set if the observation was retrieved via the 'diff' method
        :type id_: int
        :type modification_type: ModificationType
        """
        super(Observation, self).__init__(id_)
        self.modification_type = modification_type
        self._species: Optional[Species] = None
        self._observer: Optional[Observer] = None
        self._place: Optional[Place] = None
        self._form: Optional[ornitho.model.form.Form] = None
        self._resting_habitat: Optional[FieldOption] = None
        self._accuracy_of_location: Optional[FieldOption] = None
        self._observation_detail: Optional[FieldOption] = None
        self._atlas_code: Optional[FieldOption] = None
        self._medias: Optional[List[Media]] = None

    @classmethod
    def create_from_ornitho_json(cls, data: Dict[str, Any]) -> "Observation":
        if len(data["observers"]) > 1:
            raise APIException(
                f"More than one observer in sightings json found!\n{data['observers']}"
            )
        identifier: Optional[int] = (
            int(data["observers"][0]["id_sighting"])
            if "observers" in data and "id_sighting" in data["observers"][0]
            else None
        )
        obj = cls(identifier)
        obj._raw_data = data
        return obj

    @property  # type: ignore
    @check_raw_data("observers")
    def id_observer(self) -> int:
        if "id" in self._raw_data["observers"][0]:
            return int(self._raw_data["observers"][0]["id"])
        else:
            return int(self._raw_data["observers"][0]["@id"])

    @id_observer.setter
    def id_observer(self, value: int):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["@id"] = value.__str__()
        else:
            self._observer = None
            self._raw_data["observers"] = [{"@id": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def traid(self) -> int:
        return int(self._raw_data["observers"][0]["traid"])

    @property  # type: ignore
    @check_raw_data("observers")
    def guid(self) -> uuid.UUID:
        return uuid.UUID(self._raw_data["observers"][0]["guid"])

    @guid.setter
    def guid(self, value: uuid.UUID):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["guid"] = value.__str__()
        else:
            self._raw_data["observers"] = [{"guid": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def timing(self) -> datetime:
        timing = datetime.fromtimestamp(
            int(self._raw_data["observers"][0]["timing"]["@timestamp"]),
        ).astimezone()
        return timing

    @timing.setter
    def timing(self, value: datetime):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["timing"] = {
                "@timestamp": int(value.timestamp()).__str__()
            }
        else:
            self._raw_data["observers"] = [
                {"timing": {"@timestamp": int(value.timestamp()).__str__()}}
            ]
        # Add date to raw_data, so ornitho can process it
        self._raw_data["date"] = {"@timestamp": int(value.timestamp()).__str__()}

    @property  # type: ignore
    @check_raw_data("observers")
    def coord_lat(self) -> float:
        return float(self._raw_data["observers"][0]["coord_lat"])

    @coord_lat.setter
    def coord_lat(self, value: float):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["coord_lat"] = value.__str__()
        else:
            self._raw_data["observers"] = [{"coord_lat": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def coord_lon(self) -> float:
        return float(self._raw_data["observers"][0]["coord_lon"])

    @coord_lon.setter
    def coord_lon(self, value: float):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["coord_lon"] = value.__str__()
        else:
            self._raw_data["observers"] = [{"coord_lon": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def altitude(self) -> int:
        return int(self._raw_data["observers"][0]["altitude"])

    @altitude.setter
    def altitude(self, value: int):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["altitude"] = value.__str__()
        else:
            self._raw_data["observers"] = [{"altitude": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def id_form(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["id_form"])
            if "id_form" in self._raw_data["observers"][0]
            else None
        )

    @id_form.setter
    def id_form(self, value: int):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["id_form"] = value
        else:
            self._raw_data["observers"] = [{"id_form": value}]

    @property  # type: ignore
    @check_raw_data("observers")
    def precision(self) -> Precision:
        return Precision(self._raw_data["observers"][0]["precision"])

    @precision.setter
    def precision(self, value: Precision):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["precision"] = value.value
        else:
            self._raw_data["observers"] = [{"precision": value.value}]

    @property  # type: ignore
    @check_raw_data("observers")
    def estimation_code(self) -> Optional[EstimationCode]:
        return (
            EstimationCode(self._raw_data["observers"][0]["estimation_code"])
            if "estimation_code" in self._raw_data["observers"][0]
            else None
        )

    @estimation_code.setter
    def estimation_code(self, value: EstimationCode):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["estimation_code"] = value.value
        else:
            self._raw_data["observers"] = [{"estimation_code": value.value}]

    @property  # type: ignore
    @check_raw_data("species")
    def id_species(self) -> int:
        if "id" in self._raw_data["species"]:
            return int(self._raw_data["species"]["id"])
        else:
            return int(self._raw_data["species"]["@id"])

    @id_species.setter
    def id_species(self, value: int):
        self._species = None
        self._raw_data["species"] = {"@id": value.__str__()}

    @property  # type: ignore
    @check_raw_data("observers")
    def count(self) -> int:
        return int(self._raw_data["observers"][0]["count"])

    @count.setter
    def count(self, value: str):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["count"] = value.__str__()
        else:
            self._raw_data["observers"] = [{"count": value.__str__()}]

    @property  # type: ignore
    @check_raw_data("observers")
    def flight_number(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["flight_number"])
            if "flight_number" in self._raw_data["observers"][0]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def admin_hidden(self) -> bool:
        return (
            False
            if "admin_hidden" not in self._raw_data["observers"][0]
            or self._raw_data["observers"][0]["admin_hidden"] == "0"
            else True
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def admin_hidden_type(self) -> Optional[str]:
        return (
            self._raw_data["observers"][0]["admin_hidden_type"]
            if "admin_hidden_type" in self._raw_data["observers"][0]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def source(self) -> Source:
        return Source(self._raw_data["observers"][0]["source"])

    @property  # type: ignore
    @check_raw_data("observers")
    def medias(self) -> Optional[List[Media]]:
        if self._medias is None and "medias" in self._raw_data["observers"][0]:
            self._medias = [
                Media.get(media["@id"])
                for media in self._raw_data["observers"][0]["medias"]
            ]
        return self._medias

    @property  # type: ignore
    @check_raw_data("observers")
    def media_urls(self) -> Optional[List[str]]:
        if "medias" in self._raw_data["observers"][0]:
            return [
                f"{media['path']}/{media['filename']}"
                for media in self._raw_data["observers"][0]["medias"]
            ]
        return None

    @property  # type: ignore
    @check_raw_data("observers")
    def comment(self) -> Optional[str]:
        return (
            self._raw_data["observers"][0]["comment"]
            if "comment" in self._raw_data["observers"][0]
            else None
        )

    @comment.setter
    def comment(self, value: str):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["comment"] = value
        else:
            self._raw_data["observers"] = [{"comment": value}]

    @property  # type: ignore
    @check_raw_data("observers")
    def hidden_comment(self) -> Optional[str]:
        return (
            self._raw_data["observers"][0]["hidden_comment"]
            if "hidden_comment" in self._raw_data["observers"][0]
            else None
        )

    @hidden_comment.setter
    def hidden_comment(self, value: str):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["hidden_comment"] = value
        else:
            self._raw_data["observers"] = [{"hidden_comment": value}]

    @property  # type: ignore
    @check_raw_data("observers")
    def hidden(self) -> bool:
        return (
            False
            if "hidden" not in self._raw_data["observers"][0]
            or self._raw_data["observers"][0]["hidden"] == "0"
            else True
        )

    @hidden.setter
    def hidden(self, value: bool):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["hidden"] = "1" if value else "0"
        else:
            self._raw_data["observers"] = [{"hidden": "1" if value else "0"}]

    @property  # type: ignore
    @check_raw_data("observers")
    def id_atlas_code(self) -> Optional[str]:
        id_atlas_code = (
            None
            if "atlas_code" not in self._raw_data["observers"][0]
            else self._raw_data["observers"][0]["atlas_code"]["@id"]
            if type(self._raw_data["observers"][0]["atlas_code"]) is dict
            else f"3_{self._raw_data['observers'][0]['atlas_code']}"
        )
        return id_atlas_code

    @id_atlas_code.setter
    def id_atlas_code(self, value: str):
        self._atlas_code = None
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["atlas_code"] = {"@id": value}
        else:
            self._raw_data["observers"] = [{"atlas_code": {"@id": value}}]

    @property  # type: ignore
    @check_raw_data("observers")
    def atlas_code_text(self) -> Optional[str]:
        atlas_code_text = (
            self._raw_data["observers"][0]["atlas_code"]["#text"]
            if "atlas_code" in self._raw_data["observers"][0]
            and "#text" in self._raw_data["observers"][0]["atlas_code"]
            else None
        )
        return atlas_code_text

    @property  # type: ignore
    @check_raw_data("observers")
    def details(self) -> Optional[List[Detail]]:
        details = None
        if "details" in self._raw_data["observers"][0]:
            if "@id" in self._raw_data["observers"][0]["details"][0]["sex"]:
                details = [
                    Detail(
                        int(detail["count"]), detail["sex"]["@id"], detail["age"]["@id"]
                    )
                    for detail in self._raw_data["observers"][0]["details"]
                ]
            else:
                details = [
                    Detail(int(detail["count"]), detail["sex"], detail["age"])
                    for detail in self._raw_data["observers"][0]["details"]
                ]
        return details

    @details.setter
    def details(self, value: List[Detail]):
        details_ornitho_format = [
            {
                "count": detail.count.__str__(),
                "sex": {"@id": detail.sex},
                "age": {"@id": detail.age},
            }
            for detail in value
        ]
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["details"] = details_ornitho_format
        else:
            self._raw_data["observers"] = [{"details": details_ornitho_format}]

    @property  # type: ignore
    @check_raw_data("observers")
    def insert_date(self) -> datetime:
        insert_date = datetime.fromtimestamp(
            int(self._raw_data["observers"][0]["insert_date"]["@timestamp"])
            if type(self._raw_data["observers"][0]["insert_date"]) is dict
            else int(self._raw_data["observers"][0]["insert_date"]),
        ).astimezone()
        return insert_date

    @property  # type: ignore
    @check_raw_data("observers")
    def update_date(self) -> Optional[datetime]:
        update_date = (
            datetime.fromtimestamp(
                int(self._raw_data["observers"][0]["update_date"]["@timestamp"])
                if type(self._raw_data["observers"][0]["update_date"]) is dict
                else int(self._raw_data["observers"][0]["update_date"]),
            ).astimezone()
            if "update_date" in self._raw_data["observers"][0]
            else None
        )
        return update_date

    @property  # type: ignore
    @check_raw_data("place")
    def id_place(self) -> int:
        if "place" not in self._raw_data:
            id_place = Place.find_closest_place(
                self.coord_lat, self.coord_lon, get_hidden=True
            ).id_
            if id_place is not None:
                self.id_place = int(id_place)
                return int(id_place)
        if "@id" in self._raw_data["place"]:
            return int(self._raw_data["place"]["@id"])
        else:
            return int(self._raw_data["place"]["id"])

    @id_place.setter
    def id_place(self, value: int):
        self._place = None
        self._raw_data["place"] = {"@id": value.__str__()}

    @property  # type: ignore
    @check_raw_data("observers")
    def id_resting_habitat(self) -> Optional[str]:
        if "resting_habitat" in self._raw_data["observers"][0]:
            if type(self._raw_data["observers"][0]["resting_habitat"]) is dict:
                return self._raw_data["observers"][0]["resting_habitat"]["@id"]
            else:
                return self._raw_data["observers"][0]["resting_habitat"]
        return None

    @id_resting_habitat.setter
    def id_resting_habitat(self, value: str):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["resting_habitat"] = value
        else:
            self._raw_data["observers"] = [{"resting_habitat": value}]

    @property  # type: ignore
    @check_raw_data("observers")
    def id_accuracy_of_location(self) -> Optional[str]:
        if "accuracy_of_location" in self._raw_data["observers"][0]:
            if type(self._raw_data["observers"][0]["accuracy_of_location"]) is dict:
                return self._raw_data["observers"][0]["accuracy_of_location"]["@id"]
            else:
                # return self._raw_data["observers"][0]["accuracy_of_location"]
                # pass, since accuracy_of_location isn't available in short version
                pass
        return None

    @property  # type: ignore
    @check_raw_data("observers")
    def id_observation_detail(self) -> Optional[str]:
        if "observation_detail" in self._raw_data["observers"][0]:
            if type(self._raw_data["observers"][0]["observation_detail"]) is dict:
                return self._raw_data["observers"][0]["observation_detail"]["@id"]
            else:
                return self._raw_data["observers"][0]["observation_detail"]
        return None

    @id_observation_detail.setter
    def id_observation_detail(self, value: str):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["observation_detail"] = value
        else:
            self._raw_data["observers"] = [{"observation_detail": value}]

    @property  # type: ignore
    @check_raw_data("species")
    def species(self) -> Species:
        """Observed Species"""
        if self._species is None:
            if "@id" in self._raw_data["species"]:
                self._species = Species.create_from_ornitho_json(
                    self._raw_data["species"]
                )
            else:
                self._species = Species.get(self._raw_data["species"]["id"])
                self._raw_data["species"] = self._species._raw_data
        return self._species

    @species.setter
    def species(self, value: Species):
        self.id_species = value.id_
        self._species = value

    @property  # type: ignore
    @check_raw_data("observers")
    def observer(self) -> Observer:
        """Observing user"""
        if self._observer is None:
            self._observer = Observer.create_from_ornitho_json(
                self._raw_data["observers"][0]
            )
        return self._observer

    @observer.setter
    def observer(self, value: Observer):
        self.id_observer = value.id_
        self._observer = value

    @property  # type: ignore
    @check_raw_data("place")
    def place(self) -> Place:
        """Place of the observation"""
        if self._place is None:
            self._place = Place.create_from_ornitho_json(self._raw_data["place"])
        return self._place

    @place.setter
    def place(self, value: Place):
        self._raw_data["place"] = value._raw_data
        self._place = value

    @property
    def form(self):
        if self._form is None and self.id_form is not None:
            if "form" in self._raw_data:
                self._form = ornitho.model.form.Form.create_from_ornitho_json(
                    self._raw_data["form"]
                )
            else:
                self._form = ornitho.model.form.Form.get(self.id_form)
        return self._form

    @property
    def resting_habitat(self) -> Optional[FieldOption]:
        """Resting habitat of the observation"""
        if self._resting_habitat is None and self.id_resting_habitat:
            self._resting_habitat = FieldOption.get(self.id_resting_habitat)
        return self._resting_habitat

    @resting_habitat.setter
    def resting_habitat(self, value: FieldOption):
        self.id_resting_habitat = value.id_
        self._resting_habitat = value

    @property
    def accuracy_of_location(self) -> Optional[FieldOption]:
        """Resting habitat of the observation"""
        if self._accuracy_of_location is None and self.id_accuracy_of_location:
            self._accuracy_of_location = FieldOption.get(self.id_accuracy_of_location)
        return self._accuracy_of_location

    @property
    def observation_detail(self) -> Optional[FieldOption]:
        """Observation detail of the observation"""
        if self._observation_detail is None and self.id_observation_detail:
            self._observation_detail = FieldOption.get(self.id_observation_detail)
        return self._observation_detail

    @observation_detail.setter
    def observation_detail(self, value: FieldOption):
        self.id_observation_detail = value.id_
        self._observation_detail = value

    @property
    def atlas_code(self) -> Optional[FieldOption]:
        """Atlas Code of the observation"""
        if self._atlas_code is None and self.id_atlas_code:
            self._atlas_code = FieldOption.get(f"3_{self.id_atlas_code}")
        return self._atlas_code

    @atlas_code.setter
    def atlas_code(self, value: FieldOption):
        self.id_atlas_code = value.id_
        self._atlas_code = value

    @property  # type: ignore
    @check_raw_data("observers")
    def is_exported(self) -> bool:
        return (
            True
            if "observers" in self._raw_data
            and "is_exported" in self._raw_data["observers"][0]
            and self._raw_data["observers"][0]["is_exported"] == "1"
            else False
        )

    @is_exported.setter
    def is_exported(self, value: bool):
        value_as_str = "1" if value else "0"
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["is_exported"] = value_as_str
        else:
            self._raw_data["observers"] = [{"is_exported": value_as_str}]

    @property  # type: ignore
    @check_raw_data("observers")
    def export_date(self) -> Optional[datetime]:
        return (
            datetime.fromtimestamp(
                int(self._raw_data["observers"][0]["export_date"]["@timestamp"]),
            ).astimezone()
            if self.is_exported
            else None
        )

    @export_date.setter
    def export_date(self, value: datetime):
        if "observers" in self._raw_data:
            self._raw_data["observers"][0]["export_date"] = {
                "@timestamp": int(value.timestamp()).__str__()
            }
        else:
            self._raw_data["observers"] = [
                {"export_date": {"@timestamp": int(value.timestamp()).__str__()}}
            ]

    @property  # type: ignore
    @check_raw_data("observers")
    def notime(self) -> bool:
        return (
            False
            if "@notime" not in self._raw_data["observers"][0]["timing"]
            or self._raw_data["observers"][0]["timing"]["@notime"] == "0"
            else True
        )

    @notime.setter
    def notime(self, value: bool):
        if "observers" in self._raw_data:
            if "timing" in self._raw_data["observers"][0]:
                self._raw_data["observers"][0]["timing"]["@notime"] = (
                    "1" if value else "0"
                )
            else:
                self._raw_data["observers"][0]["timing"] = {
                    "@notime": "1" if value else "0"
                }
        else:
            self._raw_data["observers"] = [
                {"timing": {"@notime": "1" if value else "0"}}
            ]

    @property  # type: ignore
    @check_raw_data("observers")
    def project(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["project"])
            if "project" in self._raw_data["observers"][0]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def project_code(self) -> Optional[str]:
        return (
            self._raw_data["observers"][0]["project_code"]
            if "project_code" in self._raw_data["observers"][0]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def cavs(self) -> Optional[str]:
        return (
            self._raw_data["observers"][0]["committees_validation"]["cavs"]
            if "committees_validation" in self._raw_data["observers"][0]
            and "cavs" in self._raw_data["observers"][0]["committees_validation"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def id_observer_vowa(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["vowa_id"])
            if "vowa_id" in self._raw_data["observers"][0]
            else int(self._raw_data["observers"][0]["@vowa_id"])
            if "@vowa_id" in self._raw_data["observers"][0]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def second_hand(self) -> bool:
        return (
            False
            if "second_hand" not in self._raw_data["observers"][0]
            or self._raw_data["observers"][0]["second_hand"] == "0"
            else True
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_couples(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["extended_info"]["colony"]["couples"])
            if "extended_info" in self._raw_data["observers"][0]
            and "colony" in self._raw_data["observers"][0]["extended_info"]
            and "couples" in self._raw_data["observers"][0]["extended_info"]["colony"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_nests(self) -> Optional[int]:
        return (
            int(self._raw_data["observers"][0]["extended_info"]["colony"]["nests"])
            if "extended_info" in self._raw_data["observers"][0]
            and "colony" in self._raw_data["observers"][0]["extended_info"]
            and "nests" in self._raw_data["observers"][0]["extended_info"]["colony"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_occupied_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony"][
                    "occupied_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony" in self._raw_data["observers"][0]["extended_info"]
            and "occupied_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_nests_is_min(self) -> Optional[bool]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony"][
                    "nests_is_min"
                ]
            )
            == 1
            if "extended_info" in self._raw_data["observers"][0]
            and "colony" in self._raw_data["observers"][0]["extended_info"]
            and "nests_is_min"
            in self._raw_data["observers"][0]["extended_info"]["colony"]
            else False
            if self.colony_nests is not None
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_couples(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "couples"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "couples"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_natural_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_natural_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_natural_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_natural_nests_is_min(self) -> Optional[bool]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_natural_nests_is_min"
                ]
            )
            == 1
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_natural_nests_is_min"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else False
            if self.colony_extended_nb_natural_nests is not None
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_artificial_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_artificial_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_artificial_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_artificial_nests_is_min(self) -> Optional[bool]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_artificial_nests_is_min"
                ]
            )
            == 1
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_artificial_nests_is_min"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else False
            if self.colony_extended_nb_artificial_nests is not None
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_natural_occup_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_natural_occup_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_natural_occup_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_artificial_occup_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_artificial_occup_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_artificial_occup_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_natural_other_species_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_natural_other_species_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_natural_other_species_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_artificial_other_species_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_artificial_other_species_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_artificial_other_species_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_natural_destructed_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_natural_destructed_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_natural_destructed_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_artificial_destructed_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_artificial_destructed_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_artificial_destructed_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def colony_extended_nb_construction_nests(self) -> Optional[int]:
        return (
            int(
                self._raw_data["observers"][0]["extended_info"]["colony_extended"][
                    "nb_construction_nests"
                ]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "colony_extended" in self._raw_data["observers"][0]["extended_info"]
            and "nb_construction_nests"
            in self._raw_data["observers"][0]["extended_info"]["colony_extended"]
            else None
        )

    @property  # type: ignore
    @check_raw_data("observers")
    def nest_number(self) -> Optional[int]:
        if (
            "protocol" in self._raw_data["observers"][0]
            and "nest_number" in self._raw_data["observers"][0]["protocol"]
        ):
            if type(self._raw_data["observers"][0]["protocol"]["nest_number"]) is dict:
                return int(
                    self._raw_data["observers"][0]["protocol"]["nest_number"]["@id"]
                )
            else:
                return int(self._raw_data["observers"][0]["protocol"]["nest_number"])
        return None

    @property  # type: ignore
    @check_raw_data("observers")
    def occupied_nest_number(self) -> Optional[int]:
        if (
            "protocol" in self._raw_data["observers"][0]
            and "occupied_nest_number" in self._raw_data["observers"][0]["protocol"]
        ):
            if (
                type(self._raw_data["observers"][0]["protocol"]["occupied_nest_number"])
                is dict
            ):
                return int(
                    self._raw_data["observers"][0]["protocol"]["occupied_nest_number"][
                        "@id"
                    ]
                )
            else:
                return int(
                    self._raw_data["observers"][0]["protocol"]["occupied_nest_number"]
                )
        return None

    @property  # type: ignore
    @check_raw_data("observers")
    def potential_breeding_pairs(self) -> Optional[int]:
        if (
            "protocol" in self._raw_data["observers"][0]
            and "potential_breeding_pairs" in self._raw_data["observers"][0]["protocol"]
        ):
            if (
                type(
                    self._raw_data["observers"][0]["protocol"][
                        "potential_breeding_pairs"
                    ]
                )
                is dict
            ):
                return int(
                    self._raw_data["observers"][0]["protocol"][
                        "potential_breeding_pairs"
                    ]["@id"]
                )
            else:
                return int(
                    self._raw_data["observers"][0]["protocol"][
                        "potential_breeding_pairs"
                    ]
                )
        return None

    @property  # type: ignore
    @check_raw_data("observers")
    def relations(self) -> List[Relation]:
        return (
            [
                Relation(
                    with_id=int(relation["with"]), type=RelationType(relation["type"])
                )
                for relation in self._raw_data["observers"][0]["extended_info"][
                    "relations"
                ]
                if relation["with"] and relation["type"]
            ]
            if "extended_info" in self._raw_data["observers"][0]
            and "relations" in self._raw_data["observers"][0]["extended_info"]
            else []
        )

    @relations.setter
    def relations(self, value: List[Relation]):
        relations_ornitho_format = [
            {
                "with": str(relation.with_id),
                "type": relation.type.value,
            }
            for relation in value
        ]
        if "observers" in self._raw_data:
            if "extended_info" in self._raw_data["observers"][0]:
                self._raw_data["observers"][0]["extended_info"][
                    "relations"
                ] = relations_ornitho_format
            else:
                self._raw_data["observers"][0]["extended_info"] = {
                    "relations": relations_ornitho_format
                }
        else:
            self._raw_data["observers"] = [
                {"extended_info": {"relations": relations_ornitho_format}}
            ]

    @property  # type: ignore
    @check_raw_data("observers")
    def direction(self) -> Optional[float]:
        return (
            float(
                self._raw_data["observers"][0]["extended_info"]["direction"]["degree"]
            )
            if "extended_info" in self._raw_data["observers"][0]
            and "direction" in self._raw_data["observers"][0]["extended_info"]
            else None
        )

    @direction.setter
    def direction(self, value: float):
        if "observers" in self._raw_data:
            if "extended_info" in self._raw_data["observers"][0]:
                self._raw_data["observers"][0]["extended_info"]["direction"] = {
                    "degree": str(value)
                }
            else:
                self._raw_data["observers"][0]["extended_info"] = {
                    "direction": {"degree": str(value)}
                }
        else:
            self._raw_data["observers"] = [
                {"extended_info": {"direction": {"degree": str(value)}}}
            ]

    @classmethod
    def by_observer(
        cls,
        id_observer: int,
        pagination_key: Optional[str] = None,
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool],
    ) -> Tuple[List["Observation"], Optional[str]]:
        """Retrieves a (paged) list of observations from one observer
        :param id_observer: Current data, probably received from the API
        :param pagination_key: Pagination key, which can be used to retrieve the next page
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :param kwargs: Additional filter values
        :type id_observer: int
        :type pagination_key: Optional[str]
        :type short_version: bool
        :type retries: int
        :type kwargs: Union[str, int, float, bool]
        :return: Tuple of observations and an optional pagination key
        :rtype: Tuple[List[Observation], Optional[str]]
        """
        observations, pk = cls.list(
            request_all=False,
            pagination_key=pagination_key,
            short_version=short_version,
            id_observer=id_observer,
            retries=retries,
            **kwargs,
        )
        return observations, pk

    @classmethod
    def by_observer_all(
        cls,
        id_observer: int,
        short_version: bool = False,
        retries: int = 0,
        **kwargs: Union[str, int, float, bool],
    ) -> List["Observation"]:
        """Retrieves a list of all observations from one observer
        :param id_observer: Current data, probably received from the API
        :param short_version: Indicates, if a short version with foreign keys should be returned by the API.
        :param retries: Indicates how many retries should be performed
        :param kwargs: Additional filter values
        :type id_observer: int
        :type short_version: bool
        :type retries: int
        :type kwargs: Union[str, int, float, bool]
        :return: List of observations
        :rtype: List[Observation]
        """
        observations = cls.list_all(
            id_observer=id_observer,
            short_version=short_version,
            retries=retries,
            **kwargs,
        )
        return observations

    @classmethod
    def diff(
        cls,
        date: datetime,
        modification_type: ModificationType = None,
        id_taxo_group: int = None,
        only_protocol: Union[str, BaseModel] = None,
        only_form: bool = None,
        retrieve_observations: bool = False,
        retries: int = 0,
    ) -> Dict[ModificationType, Union[List["Observation"], List[int]]]:
        """Retrieves observations which changed in between now and a given date
        :param date: Date in the past, to which changed observation should be searched
        :param modification_type: Type of modification.
        :param id_taxo_group: Optional taxo group, to which the observerd species must belong to
        :param only_protocol: Return only observation which are part of the given Protocol (Protocol Instance or Name)
        :param only_form: Return only observation which are part of a form
        :param retrieve_observations: Indicates if the observation object should be retrieved. Default: False
        :param retries: Indicates how many retries should be performed
        :type date: datetime
        :type modification_type: ModificationType
        :type id_taxo_group: int
        :type only_protocol: Union[str, "Protocol"]
        :type only_form: bool
        :type retrieve_observations: bool
        :type retries: int
        :return: Dictionary containing List of Ids/Observations as values and ModificationType as key
        :rtype: Dict[ModificationType, Union[List["Observation"], List[int]]]
        """
        url = f"{cls.ENDPOINT}/diff"
        params = dict()
        if modification_type:
            params["modification_type"] = modification_type.value
        if id_taxo_group:
            params["id_taxo_group"] = id_taxo_group.__str__()
        if only_protocol:
            from ornitho.model.protocol import Protocol

            if isinstance(only_protocol, Protocol):
                params["only_protocol"] = only_protocol.name
            else:
                params["only_protocol"] = only_protocol.__str__()
        if only_form:
            params["only_form"] = "1"

        date = date.replace(microsecond=0)
        if date.tzinfo:
            date = date.astimezone(datetime.now().astimezone().tzinfo).replace(
                tzinfo=None
            )
        params[
            "date"
        ] = (
            date.isoformat()
        )  # Format here, because isoformat is mostly ignored, except here

        changed_observations = cls.request(
            method="get", url=url, params=params, retries=retries
        )

        updated_observations = []
        updated_observations_id = []
        deleted_observations_id = []
        for obs in changed_observations:
            if obs["modification_type"] == "updated":
                updated_observations_id.append(int(obs["id_sighting"]))
            else:
                deleted_observations_id.append(int(obs["id_sighting"]))

        if retrieve_observations:
            observations_chunks = [
                updated_observations_id[i : i + 1000]
                for i in range(0, len(updated_observations_id), 1000)
            ]
            for observations_chunk in observations_chunks:
                updated_observations += cls.search(
                    period_choice="all", id_sightings_list=observations_chunk
                )[0]
        else:
            updated_observations = [
                cls(id_=obs_id) for obs_id in updated_observations_id
            ]

        return {
            ModificationType.ONLY_MODIFIED: updated_observations,
            ModificationType.ONLY_DELETED: deleted_observations_id,
        }

    @classmethod
    def create(  # type: ignore
        cls,
        observer: Union[int, Observer],
        species: Union[int, Species],
        timing: datetime,
        coord_lat: float,
        coord_lon: float,
        precision: Precision,
        estimation_code: EstimationCode,
        guid: uuid.UUID = None,
        id_form: int = None,
        place: Optional[Union[int, Place]] = None,
        notime: bool = False,
        count: int = None,
        altitude: int = None,
        comment: str = None,
        hidden_comment: str = None,
        hidden: bool = False,
        export_date: datetime = None,
        atlas_code: Union[str, FieldOption] = None,
        details: List[Detail] = None,
        resting_habitat: Union[str, FieldOption] = None,
        observation_detail: Union[str, FieldOption] = None,
        # relations: List[Relation] = None,
        # direction: float = None,
        create_in_ornitho: bool = True,
        retries: int = 0,
    ) -> "Observation":
        observation = cls()

        if isinstance(observer, Observer):
            observation.observer = observer
        else:
            observation.id_observer = observer

        if isinstance(species, Species):
            observation.species = species
        else:
            observation.id_species = species

        if guid:
            observation.guid = guid

        if id_form:
            observation.id_form = id_form

        if place:
            if isinstance(place, Place):
                observation.place = place
            else:
                observation.id_place = place

        observation.timing = timing
        observation.notime = notime
        observation.coord_lat = coord_lat
        observation.coord_lon = coord_lon
        observation.precision = precision
        observation.estimation_code = estimation_code
        if count:
            observation.count = count
        if altitude:
            observation.altitude = altitude
        if comment:
            observation.comment = comment

        if hidden_comment:
            observation.hidden_comment = hidden_comment

        if hidden:
            observation.hidden = hidden

        if export_date:
            observation.is_exported = True
            observation.export_date = (
                export_date  # Doesn't work as expected, will always be 1970-01-01 ...
            )

        if atlas_code:
            if isinstance(atlas_code, FieldOption):
                observation.atlas_code = atlas_code
            else:
                observation.id_atlas_code = atlas_code

        if details:
            observation.details = details

        if resting_habitat:
            if isinstance(resting_habitat, FieldOption):
                observation.resting_habitat = resting_habitat
            else:
                observation.id_resting_habitat = resting_habitat

        if observation_detail:
            if isinstance(observation_detail, FieldOption):
                observation.observation_detail = observation_detail
            else:
                observation.id_observation_detail = observation_detail

        # if relations:
        #     observation.relations = relations
        #
        # if direction is not None:
        #     observation.direction = direction

        if create_in_ornitho:
            observation._id = cls.create_in_ornitho(
                data={"sightings": [observation.raw_data_trim_field_ids()]},
                retries=retries,
            )
        return observation

    def mark_as_exported(self, export_date: Optional[datetime] = None):
        if self.id_ is None:
            self.refresh()
        self.is_exported = True
        if export_date:
            self.export_date = export_date
        else:
            self.export_date = datetime.now() - timedelta(
                seconds=2
            )  # Subtract 2 seconds, because the ornitho time is a little ahead
        self.update()

    def raw_data_trim_field_ids(self) -> Dict[str, Any]:
        raw_data = deepcopy(self._raw_data)

        if self.id_resting_habitat:
            raw_data["observers"][0]["resting_habitat"] = self.id_resting_habitat.split(
                "_"
            )[1]

        if self.id_observation_detail:
            raw_data["observers"][0][
                "observation_detail"
            ] = self.id_observation_detail.split("_")[1]

        if self._id:
            raw_data["observers"][0]["id_sighting"] = str(self._id)

        return raw_data

    def update_direction(
        self,
        direction: int,
        retries: int = 0,
    ):
        url = f"{self.ENDPOINT}/direction/{self.id_}"
        self.request(
            method="PUT", url=url, params={"direction": direction}, retries=retries
        )
        self.direction = direction

    def add_relation(
        self,
        with_id: int,
        type: RelationType,
        retries: int = 0,
    ):
        url = f"{self.ENDPOINT}/relations/{self.id_}"
        self.request(
            method="PUT",
            url=url,
            params={"with": with_id, "type": type.value},
            retries=retries,
        )
        self.relations = self.relations + [Relation(with_id=with_id, type=type)]



from typing import Optional

from ornitho.model.abstract import ListableModel
from ornitho.model.taxo_group import TaxonomicGroup


class Family(ListableModel):

    ENDPOINT: str = "families"

    def __init__(self, id_: int) -> None:
        """Family constructor
        :param id_: ID, which is used to get the observation from Biolovison
        :type id_: int
        """
        super(Family, self).__init__(id_)
        self._taxo_group: Optional[TaxonomicGroup] = None

    @property
    def id_taxo_group(self) -> int:
        return int(self._raw_data["id_taxo_group"])

    @property
    def name(self) -> str:
        return self._raw_data["name"]

    @property
    def latin_name(self) -> str:
        return self._raw_data["latin_name"]

    @property
    def generic(self) -> bool:
        return False if self._raw_data.get("generic") == "0" else True

    @property
    def taxo_group(self) -> TaxonomicGroup:
        if self._taxo_group is None:
            self._taxo_group = TaxonomicGroup.get(self.id_taxo_group)
        return self._taxo_group



from copy import deepcopy
from datetime import date, time
from typing import Any, Dict, List, Optional, Union

import ornitho.model.observation
from ornitho.api_exception import ObjectNotFoundException
from ornitho.api_requester import APIRequester
from ornitho.model.abstract import CreateableModel, DeletableModel
from ornitho.model.observation import Observation
from ornitho.model.place import Place
from ornitho.model.protocol import Protocol
from ornitho.model.species import Species


class Form(CreateableModel, DeletableModel):
    ENDPOINT = "observations/search"
    CREATE_ENDPOINT = "observations"
    DELETE_METHOD = "POST"
    DELETE_ENDPOINT = "observations/delete_list"

    def __init__(self, id_: int = None) -> None:
        """Form constructor
        :param id_: ID, which is used to get the form from Biolovison
        :type id_: int
        """
        super(Form, self).__init__(id_)
        self._observations: Optional[List[Observation]] = None
        self._id_place: Optional[Union[str, int]] = None

    def instance_url(self) -> str:
        """Returns url for this instance
        :return: Instance's url
        :rtype: str
        """
        return f"{self.ENDPOINT}"

    def refresh(
        self,
        short_version: bool = False,
        retries: int = 0,
    ) -> "Form":
        """Refresh local model
        Call the api and refresh fields from response
        :return: Refreshed Object
        :rtype: Form
        """
        with APIRequester() as requester:
            data, pagination_key = requester.request_raw(
                method="POST",
                url=self.instance_url(),
                short_version=short_version,
                body={"id_form": self.id_},
                retries=retries,
            )
            if "data" in data and "forms" in data["data"]:
                data = data["data"]["forms"][0]
                self._previous = self._raw_data
                self._raw_data = data
                self._observations = None
            else:
                raise ObjectNotFoundException(f"Get {data} for {self.instance_url()}")
        return self

    @property
    def id_form_universal(self) -> str:
        return self._raw_data["id_form_universal"]

    @property
    def day(self) -> Optional[date]:
        if "day" in self._raw_data:
            return date.fromtimestamp(
                int(self._raw_data["day"]["@timestamp"]),
            )
        elif self.observations is not None:
            return self.observations[0].timing.date()
        else:
            return None

    @property
    def time_start(self) -> time:
        splitted = self._raw_data["time_start"].split(":")
        return time(
            hour=int(splitted[0]),
            minute=int(splitted[1]),
            second=int(splitted[2]),
        )

    @time_start.setter
    def time_start(self, value: time):
        self._raw_data["time_start"] = value.strftime("%H:%M:%S")

    @property
    def time_stop(self) -> time:
        splitted = self._raw_data["time_stop"].split(":")
        return time(
            hour=int(splitted[0]),
            minute=int(splitted[1]),
            second=int(splitted[2]),
        )

    @time_stop.setter
    def time_stop(self, value: time):
        self._raw_data["time_stop"] = value.strftime("%H:%M:%S")

    @property
    def full_form(self) -> bool:
        return False if self._raw_data.get("full_form") == "0" else True

    @full_form.setter
    def full_form(self, value: bool):
        self._raw_data["full_form"] = "1" if value else "0"

    @property
    def version(self) -> int:
        return int(self._raw_data["version"])

    @property
    def lat(self) -> float:
        return float(self._raw_data["lat"])

    @property
    def lon(self) -> float:
        return float(self._raw_data["lon"])

    @property
    def id_form_mobile(self) -> Optional[str]:
        return (
            self._raw_data["id_form_mobile"]
            if "id_form_mobile" in self._raw_data
            else None
        )

    @property
    def comment(self) -> Optional[str]:
        return self._raw_data["comment"] if "comment" in self._raw_data else None

    @comment.setter
    def comment(self, value: str):
        self._raw_data["comment"] = value

    @property
    def protocol_name(self) -> Optional[str]:
        return (
            self._raw_data["protocol"]["protocol_name"]
            if "protocol" in self._raw_data
            else None
        )

    @protocol_name.setter
    def protocol_name(self, value: str):
        if "protocol" in self._raw_data:
            self._raw_data["protocol"]["protocol_name"] = value
        else:
            self._raw_data["protocol"] = {"protocol_name": value}

    @property
    def site_code(self) -> Optional[str]:
        return (
            self._raw_data["protocol"]["site_code"]
            if "protocol" in self._raw_data
            else None
        )

    @site_code.setter
    def site_code(self, value: str):
        if "protocol" in self._raw_data:
            self._raw_data["protocol"]["site_code"] = value
        else:
            self._raw_data["protocol"] = {"site_code": value}

    @property
    def local_site_code(self) -> Optional[str]:
        return (
            self._raw_data["protocol"]["local_site_code"]
            if "protocol" in self._raw_data
            else None
        )

    @property
    def advanced(self) -> Optional[bool]:
        return (
            self._raw_data["protocol"]["advanced"] != "0"
            if "protocol" in self._raw_data
            else None
        )

    @property
    def visit_number(self) -> Optional[int]:
        return (
            int(self._raw_data["protocol"]["visit_number"])
            if "protocol" in self._raw_data
            else None
        )

    @visit_number.setter
    def visit_number(self, value: int):
        if "protocol" in self._raw_data:
            self._raw_data["protocol"]["visit_number"] = value.__str__()
        else:
            self._raw_data["protocol"] = {"visit_number": value.__str__()}

    @property
    def sequence_number(self) -> Optional[int]:
        return (
            int(self._raw_data["protocol"]["sequence_number"])
            if "protocol" in self._raw_data
            else None
        )

    @sequence_number.setter
    def sequence_number(self, value: int):
        if "protocol" in self._raw_data:
            self._raw_data["protocol"]["sequence_number"] = value.__str__()
        else:
            self._raw_data["protocol"] = {"sequence_number": value.__str__()}

    @property
    def list_type(self) -> Optional[str]:
        return (
            self._raw_data["protocol"]["list_type"]
            if "protocol" in self._raw_data
            else None
        )

    @property
    def wkt(self) -> Optional[str]:
        """
        Well-known text representation of the recorded track
        :return: LINESTRING geometry, if a valid LINESTRING WKT is received, None otherwise
        """
        return (
            self._raw_data["protocol"]["wkt"]
            if "protocol" in self._raw_data
            and "wkt" in self._raw_data["protocol"]
            and "LINESTRING" in self._raw_data["protocol"]["wkt"]  # check if Linestring
            and ","
            in self._raw_data["protocol"][
                "wkt"
            ]  # check if Linestring contains 2 points
            else None
        )

    @property
    def id_waterbird_conditions(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_conditions" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_conditions"]) is dict:
                return self._raw_data["protocol"]["waterbird_conditions"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_conditions"].split(",")[0]
        return None

    @id_waterbird_conditions.setter
    def id_waterbird_conditions(self, value: str):
        if "protocol" in self._raw_data:
            self._raw_data["protocol"]["waterbird_conditions"] = {"@id": value}
        else:
            self._raw_data["protocol"] = {"waterbird_conditions": {"@id": value}}

    @property
    def id_waterbird_coverage(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_coverage" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_coverage"]) is dict:
                return self._raw_data["protocol"]["waterbird_coverage"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_coverage"].split(",")[0]
        return None

    @property
    def id_waterbird_optical(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_optical" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_optical"]) is dict:
                return self._raw_data["protocol"]["waterbird_optical"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_optical"].split(",")[0]
        return None

    @property
    def id_waterbird_countmethod(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_countmethod" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_countmethod"]) is dict:
                return self._raw_data["protocol"]["waterbird_countmethod"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_countmethod"].split(",")[0]
        return None

    @property
    def id_waterbird_ice(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_ice" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_ice"]) is dict:
                return self._raw_data["protocol"]["waterbird_ice"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["waterbird_ice"].split(",")[0]
        return None

    @property
    def id_waterbird_snowcover(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_snowcover" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_snowcover"]) is dict:
                return self._raw_data["protocol"]["waterbird_snowcover"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_snowcover"].split(",")[0]
        return None

    @property
    def id_waterbird_waterlevel(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_waterlevel" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_waterlevel"]) is dict:
                return self._raw_data["protocol"]["waterbird_waterlevel"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_waterlevel"].split(",")[0]
        return None

    @property
    def id_waterbird_counttype(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_counttype" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_counttype"]) is dict:
                return self._raw_data["protocol"]["waterbird_counttype"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_counttype"].split(",")[0]
        return None

    @property
    def id_waterbird_visibility(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_visibility" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_visibility"]) is dict:
                return self._raw_data["protocol"]["waterbird_visibility"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_visibility"].split(",")[0]
        return None

    @property
    def id_waterbird_waves(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_waves" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_waves"]) is dict:
                return self._raw_data["protocol"]["waterbird_waves"]["@id"].split(",")[
                    0
                ]
            else:
                return self._raw_data["protocol"]["waterbird_waves"].split(",")[0]
        return None

    @property
    def id_waterbird_conditions_reason(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_conditions_reason" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_conditions_reason"]) is dict:
                return self._raw_data["protocol"]["waterbird_conditions_reason"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"]["waterbird_conditions_reason"].split(
                    ","
                )[0]
        return None

    @property
    def id_waterbird_count_payed(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_count_payed" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_count_payed"]) is dict:
                return self._raw_data["protocol"]["waterbird_count_payed"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["waterbird_count_payed"].split(",")[0]
        return None

    @property
    def id_waterbird_activity_persons_on_shore(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_persons_on_shore" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_persons_on_shore"])
                is dict
            ):
                return self._raw_data["protocol"][
                    "waterbird_activity_persons_on_shore"
                ]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_persons_on_shore"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_boats_rowing(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_boats_rowing" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_boats_rowing"])
                is dict
            ):
                return self._raw_data["protocol"]["waterbird_activity_boats_rowing"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_boats_rowing"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_boats_motor(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_boats_motor" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_boats_motor"])
                is dict
            ):
                return self._raw_data["protocol"]["waterbird_activity_boats_motor"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_boats_motor"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_boats_sailing(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_boats_sailing" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_boats_sailing"])
                is dict
            ):
                return self._raw_data["protocol"]["waterbird_activity_boats_sailing"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_boats_sailing"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_boats_kayak(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_boats_kayak" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_boats_kayak"])
                is dict
            ):
                return self._raw_data["protocol"]["waterbird_activity_boats_kayak"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_boats_kayak"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_boats_fisherman(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_boats_fisherman" in self._raw_data["protocol"]
        ):
            if (
                type(self._raw_data["protocol"]["waterbird_activity_boats_fisherman"])
                is dict
            ):
                return self._raw_data["protocol"]["waterbird_activity_boats_fisherman"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"][
                    "waterbird_activity_boats_fisherman"
                ].split(",")[0]
        return None

    @property
    def id_waterbird_activity_divers(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_divers" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_activity_divers"]) is dict:
                return self._raw_data["protocol"]["waterbird_activity_divers"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"]["waterbird_activity_divers"].split(
                    ","
                )[0]
        return None

    @property
    def id_waterbird_activity_surfers(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "waterbird_activity_surfers" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["waterbird_activity_surfers"]) is dict:
                return self._raw_data["protocol"]["waterbird_activity_surfers"][
                    "@id"
                ].split(",")[0]
            else:
                return self._raw_data["protocol"]["waterbird_activity_surfers"].split(
                    ","
                )[0]
        return None

    @property
    def id_moving_harvest(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "moving_harvest" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["moving_harvest"]) is dict:
                return self._raw_data["protocol"]["moving_harvest"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["moving_harvest"].split(",")[0]
        return None

    @property
    def id_coverage(self) -> Optional[str]:
        if "protocol" in self._raw_data and "coverage" in self._raw_data["protocol"]:
            if type(self._raw_data["protocol"]["coverage"]) is dict:
                return self._raw_data["protocol"]["coverage"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["coverage"].split(",")[0]
        return None

    @property
    def id_condition(self) -> Optional[str]:
        if "protocol" in self._raw_data and "condition" in self._raw_data["protocol"]:
            if type(self._raw_data["protocol"]["condition"]) is dict:
                return self._raw_data["protocol"]["condition"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["condition"].split(",")[0]
        return None

    @property
    def id_chiro_identify(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "chiro_identify" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["chiro_identify"]) is dict:
                return self._raw_data["protocol"]["chiro_identify"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["chiro_identify"].split(",")[0]
        return None

    @property
    def id_additional_observer(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "additional_observer" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["additional_observer"]) is dict:
                return self._raw_data["protocol"]["additional_observer"]["@id"].split(
                    ","
                )[0]
            else:
                return self._raw_data["protocol"]["additional_observer"].split(",")[0]
        return None

    @property
    def id_changes(self) -> Optional[str]:
        if "protocol" in self._raw_data and "changes" in self._raw_data["protocol"]:
            if type(self._raw_data["protocol"]["changes"]) is dict:
                return self._raw_data["protocol"]["changes"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["changes"].split(",")[0]
        return None

    @property
    def id_drone_used(self) -> Optional[str]:
        if "protocol" in self._raw_data and "drone_used" in self._raw_data["protocol"]:
            if type(self._raw_data["protocol"]["drone_used"]) is dict:
                return self._raw_data["protocol"]["drone_used"]["@id"].split(",")[0]
            else:
                return self._raw_data["protocol"]["drone_used"].split(",")[0]
        return None

    @property
    def id_tmp_water_bodies(self) -> Optional[str]:
        if (
            "protocol" in self._raw_data
            and "tmp_water_bodies" in self._raw_data["protocol"]
        ):
            if type(self._raw_data["protocol"]["tmp_water_bodies"]) is dict:
                return self._raw_data["protocol"]["tmp_water_bodies"]["@id"].split(",")[
                    0
                ]
            else:
                return self._raw_data["protocol"]["tmp_water_bodies"].split(",")[0]
        return None

    @property
    def playbacks(self) -> Optional[Dict[int, bool]]:
        if "protocol" in self._raw_data and "playback" in self._raw_data["protocol"]:
            species_ids: Dict[int, bool] = {}
            for key, value in self._raw_data["protocol"]["playback"].items():
                species_id = key.replace("Id_species_", "")
                species_ids[int(species_id)] = value == "1"
            return species_ids
        return None

    @property
    def observations(self) -> Optional[List[Observation]]:
        if "sightings" not in self._raw_data:
            self.refresh()
        if self._observations is None and "sightings" in self._raw_data:
            self._observations = [
                ornitho.model.observation.Observation.create_from_ornitho_json(
                    observation
                )
                for observation in self._raw_data["sightings"]
            ]

        return self._observations

    @observations.setter
    def observations(self, value: List[Observation]):
        for observation in value:
            if self._id_place is not None:
                observation.id_place = self._id_place
            else:
                self._id_place = observation.id_place
        self._raw_data["sightings"] = [observation._raw_data for observation in value]
        self._observations = value

    def playblack_played(self, species: Union[int, Species]) -> Optional[bool]:
        if isinstance(species, Species):
            species = int(species.id_) if species.id_ is not None else 0
        return (
            self.playbacks[species]
            if self.playbacks is not None and species in self.playbacks
            else None
        )

    @classmethod
    def create(  # type: ignore
        cls,
        time_start: time,
        time_stop: time,
        observations: List[Observation],
        protocol: Optional[Union[Protocol, str]] = None,
        comment: str = None,
        place: Optional[Union[Place, int]] = None,
        visit_number: Optional[int] = None,
        sequence_number: Optional[int] = None,
        full_form: bool = True,
        protocol_headers: Dict[str, Union[int, str]] = {},
        create_in_ornitho: bool = True,
        chunk_size: int = 128,
        retries: int = 0,
    ) -> "Form":
        form = cls()
        form.time_start = time_start
        form.time_stop = time_stop
        form.full_form = full_form

        if comment is not None:
            form.comment = comment

        if place is not None:
            if isinstance(place, Place):
                form._id_place = place.id_
            else:
                form._id_place = place

        if protocol is not None:
            if isinstance(protocol, Protocol):
                form.protocol_name = protocol.name
            else:
                form.protocol_name = protocol
            for key, value in protocol_headers.items():
                if key.startswith("id_"):
                    form._raw_data["protocol"][key.lstrip("id_")] = {"@id": str(value)}
                else:
                    form._raw_data["protocol"][key] = value

        if visit_number is not None:
            form.visit_number = visit_number
        if sequence_number is not None:
            form.sequence_number = sequence_number

        if create_in_ornitho:
            # Form must be created first,
            # otherwise ornitho will ignore the given GUID/UUID for observations (╯°□°）╯︵ ┻━┻)

            # Create Form with an "Keine Art" observation, to create form in ornitho
            form.observations = [
                Observation.create(
                    observer=observations[0].id_observer,
                    species=10000,
                    place=place,
                    timing=observations[0].timing,
                    coord_lat=observations[0].coord_lat,
                    coord_lon=observations[0].coord_lon,
                    precision=observations[0].precision,
                    estimation_code=ornitho.EstimationCode.EXACT_VALUE,
                    count=0,
                    create_in_ornitho=False,
                    retries=retries,
                )
            ]
            first_observation_id = cls.create_in_ornitho(
                data={"forms": [form.raw_data_trim_field_ids()]}, retries=retries
            )
            form_id = Observation.get(first_observation_id).id_form

            # Create observations with form id in ornitho
            try:
                # Add form id to every observation
                for observation in observations:
                    observation.id_form = form_id

                for observation_chunk in [
                    observations[i : i + chunk_size]
                    for i in range(0, len(observations), chunk_size)
                ]:
                    Observation.create_in_ornitho(
                        data={
                            "sightings": [
                                observation.raw_data_trim_field_ids()
                                for observation in observation_chunk
                            ]
                        },
                        retries=retries,
                    )
                # Retrieve form again, to get acces to observation ids
                form = cls.get(form_id)
            except Exception as ex:
                form = cls(id_=form_id)
                form.delete()
                raise ex
        else:
            form.observations = observations
        return form

    def raw_data_trim_field_ids(self) -> Dict[str, Any]:
        raw_data = deepcopy(self._raw_data)
        if self.observations:
            trimmed_observations = [
                observation.raw_data_trim_field_ids()
                for observation in self.observations
            ]
            raw_data["sightings"] = trimmed_observations
        return raw_data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ornitho', 'ornitho.model', 'ornitho.model.abstract']

package_data = \
{'': ['*']}

install_requires = \
['pytz', 'requests-oauthlib']

extras_require = \
{':extra == "caching"': ['requests-cache>=0.9.3,<0.10.0']}

setup_kwargs = {
    'name': 'ornitho',
    'version': '0.3.0',
    'description': 'An ornitho API client',
    'long_description': 'Python Ornitho API Client\n=========================\n\nThe following "ornitho controllers / calls" are implemented:\n\n- taxonomic groups\n\n  - List taxo groups\n  - Get a single taxo group\n\n- families\n\n  - List families\n  - Get a single family\n\n- species\n\n  - List species\n  - Get a single species\n\n- territorial units\n\n  - List territorial units\n  - Get a single territorial unit\n\n- local admin units\n\n  - List local admin units\n  - Get a single local admin unit\n\n- places\n\n  - List places\n  - Get a single place\n\n- observers\n\n  - List Observers\n  - Get a single observer\n  - Get an observer rights list\n  - Get current observers info\n\n- entities\n\n  - List entities\n  - Get a single entity\n\n- protocol\n\n  - List protocol\n  - Get a single protocol\n  - Get list of sites for protocol id\n  - Get PDF for site id\n\n- observations\n\n  - List observations\n  - Get a single observation\n  - Create an observation\n  - Update an observations\n  - Delete an observations\n  - Search an observation\n  - Search update or deletion since a given date\n  - Delete a record\n  - Delete a list\n\n- fields\n\n  - List fields\n  - Get a single field options\n\n- media\n\n  - Get a single media\n\nNot yet Implemented:\n\n- local admin units\n\n  - Search local admin units\n\n- export organizations\n\n  - List export organizations\n  - Get a single export organization\n\n- observations\n\n  - Search closest observations\n  - Search last observations\n  - Search rare observations\n  - Search my observations\n  - Search observations from a list of observers\n  - Search observations from a list of coordinates\n  - Search resources linked to an observation\n  - Create a new resources for an observation\n\n- import files\n\n  - List import files\n  - Get a single import file\n\n- import files/observations\n\n  - List links for import files and observations\n  - Get a single link between import file and observation\n\n- validations\n\n  - List validations\n  - Get a single validation\n\n- mortality information\n\n  - List mortality informations\n\n- bearded vulture birds\n\n  - List all bearded vulture birds\n  - Get a single bird\n\n- bearded vulture information\n\n  - List bearded vulture informations\n\n- observations by polygon\n\n  - List observations_by_polygon\n  - Get a single family\n  - Get a single group\n\n- polygons\n\n  - Get a single polygon\n  - List polygons from a group\n  - List polygon groups\n  - List cache for a given polygon or group with a period\n  - List form cache for a given polygon or group with a period\n  - List species form cache for a given polygon or group with a period\n  - List observer cache for a given polygon or group\n  - List observer species cache for a given polygon or group\n\nInstalling\n----------\nThe project is published on `PyPI <https://pypi.python.org/pypi/ornitho>`__, and can be install using pip or any other\npackage manager:\n\n``$ pip install ornitho``\n\nManual installation can be done with following command:\n\n``$ pip install .``\n\nOptional caching dependency can be installed with\n\n``$ pip install ornitho[caching]``\n\nUsage\n-----\nBefore the client can be used  **consumer_key**, **consumer_secret**, **user_email**, **user_pw** and **api_base** must\nbe set:\n\n.. code-block:: python\n\n    import ornitho\n\n    ornitho.consumer_key = "CONSUMER_KEY"\n    ornitho.consumer_secret = "CONSUMER_SECRET"\n    ornitho.user_email = "USER_MAIL"\n    ornitho.user_pw = "USER_PASSWORD"\n    ornitho.api_base = "https://www.ornitho.de/api/"\n\nThe client can then be used.\n\nCaching\n~~~~~~~~\nIf the additional caching dependency is installed, non chunked responses can be cached.\nFor activating the cache one has to set the `cache_enabled` flag.\nFollowing settings are available:\n\n.. code-block:: python\n\n    ornitho.cache_enabled = True          # Enable/Disable caching\n    ornitho.cache_name = "ornitho_cache"  # Name of the cache\n    ornitho.cache_backend = "redis"       # Set caching backend, possible values: sqlite, memory, mongodb, redis\n    ornitho.cache_expire_after = 600      # Set expiration time for cached responses\n\nFor more information see the `requests-cache documentation\n<https://requests-cache.readthedocs.io/en/latest/index.html>`__\n\nExamples\n~~~~~~~~\nFollowing code shows how to get all observation from ornitho.de between 01.10.2019 and 31.10.2019:\n\n.. code-block:: python\n\n    import os\n    import ornitho\n\n    ornitho.consumer_key = os.environ.get("ORNITHO_CONSUMER_KEY")\n    ornitho.consumer_secret = os.environ.get("ORNITHO_CONSUMER_SECRET")\n    ornitho.user_email = os.environ.get("ORNITHO_USER_EMAIL")\n    ornitho.user_pw = os.environ.get("ORNITHO_USER_PW")\n    ornitho.api_base = "https://www.ornitho.de/api/"\n    \n    resp = ornitho.Observation.search_all(period_choice="range", date_from="01.10.2019", date_to="31.10.2019")\n    print(f"Found {len(resp)} observations between 01.10.2019 and 31.10.2019")\n    \nMore examples can be found the `examples <https://github.com/dda-dev/ornitho-client-python/tree/master/examples>`__\nfolder.\n\nPrerequisites\n~~~~~~~~~~~~~\n\nThe project has been tested with the following python versions:\n\n- Python 3.7\n- Python 3.8\n- Python 3.9\n- Python 3.10\n\nCollaborate\n-----------\n\nAny kind of help with the project will be well received, and there are\ntwo main ways to give such help:\n\n- Reporting errors and asking for extensions through the issues management\n- or forking the repository and extending the project\n\nIssues management\n~~~~~~~~~~~~~~~~~\n\nIssues are managed at the Github `project issues\ntracker <https://github.com/dda-dev/ornitho-client-python/issues>`__, where\nany Github user may report bugs or ask for new features.\n\nTesting\n~~~~~~~\n\nThe tests included with the project can be run with:\n\n``$ pytest``\n\nor with a coverage report:\n\n``$ pytest --cov=ornitho tests/``\n\nTo test all supported Python versions, use tox:\n\n``$ tox``\n\nLicense\n-------\n\nThe project has been released under the `MIT\nLicense <https://opensource.org/licenses/MIT>`__.\n',
    'author': 'Patrick Lindel',
    'author_email': 'lindel@dda-web.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dda-dev/ornitho-client-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# flake8: noqa

from __future__ import absolute_import, division, print_function, unicode_literals

from ornitho.model.abstract.base_model import BaseModel
from ornitho.model.abstract.createable_model import CreateableModel
from ornitho.model.abstract.deletable_model import DeletableModel
from ornitho.model.abstract.listable_model import ListableModel
from ornitho.model.abstract.searchable_model import SearchableModel
from ornitho.model.abstract.updateable_model import UpdateableModel



from abc import ABC
from typing import Optional, TypeVar

from ornitho.model.abstract import BaseModel

# Create a generic variable that can be 'DeletableModel', or any subclass.
T = TypeVar("T", bound="DeletableModel")


class DeletableModel(BaseModel, ABC):
    """Abstract class for deletable models via DELETE /ENDPOINT"""

    DELETE_METHOD: str = "DELETE"
    DELETE_ENDPOINT: Optional[str] = None

    def delete(
        self: T,
        retries: int = 0,
    ):
        """Delete an instance on ornitho"""
        url = f"{self.DELETE_ENDPOINT if self.DELETE_ENDPOINT is not None else self.ENDPOINT}/{self.id_}"
        self.request(method=self.DELETE_METHOD, url=url, retries=retries)



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from ornitho.model.abstract import ListableModel
from ornitho.model.abstract.base_model import BaseModel, check_refresh
from ornitho.model.local_admin_unit import LocalAdminUnit
from ornitho.model.modification_type import ModificationType
from ornitho.model.observer import Observer


class Place(ListableModel):
    ENDPOINT: str = "places"

    def __init__(self, id_: int, modification_type: ModificationType = None) -> None:
        super(Place, self).__init__(id_)
        self.modification_type = modification_type
        self._local_admin_unit: Optional[LocalAdminUnit] = None
        self._centroid: Optional[str] = None
        self._order: Optional[int] = None
        self._wkt: Optional[str] = None

    @property  # type: ignore
    @check_refresh
    def id_commune(self) -> int:
        """ID, in which the place is located"""
        return int(self._raw_data["id_commune"])

    @property  # type: ignore
    @check_refresh
    def name(self) -> str:
        return self._raw_data["name"]

    @property  # type: ignore
    @check_refresh
    def coord_lon(self) -> float:
        return (
            float(self._raw_data["coord_lon"])
            if "coord_lon" in self._raw_data
            else self._raw_data["lon"]
        )

    @property  # type: ignore
    @check_refresh
    def coord_lat(self) -> float:
        return (
            float(self._raw_data["coord_lat"])
            if "coord_lat" in self._raw_data
            else self._raw_data["lat"]
        )

    @property  # type: ignore
    @check_refresh
    def altitude(self) -> int:
        return int(self._raw_data["altitude"])

    @property  # type: ignore
    @check_refresh
    def id_region(self) -> int:
        return int(self._raw_data["id_region"])

    @property  # type: ignore
    @check_refresh
    def visible(self) -> bool:
        return False if self._raw_data.get("visible") == "0" else True

    @property  # type: ignore
    @check_refresh
    def is_private(self) -> bool:
        return False if self._raw_data.get("is_private") == "0" else True

    @property  # type: ignore
    @check_refresh
    def place_type(self) -> str:
        return self._raw_data["place_type"]

    @property  # type: ignore
    @check_refresh
    def loc_precision(self) -> int:
        return int(self._raw_data["loc_precision"])

    @property
    def local_admin_unit(self) -> LocalAdminUnit:
        if self._local_admin_unit is None:
            self._local_admin_unit = LocalAdminUnit(id_=self.id_commune)
        return self._local_admin_unit

    @property
    def commune(self) -> LocalAdminUnit:
        return self.local_admin_unit

    @property  # type: ignore
    @check_refresh
    def created_by(self) -> Observer:
        return Observer(id_=int(self._raw_data["created_by"]))

    @property  # type: ignore
    @check_refresh
    def created_date(self) -> datetime:
        created_date = datetime.fromtimestamp(
            int(self._raw_data["created_date"]["@timestamp"])
            if type(self._raw_data["created_date"]) is dict
            else int(self._raw_data["created_date"]),
        ).astimezone()
        return created_date

    @property  # type: ignore
    @check_refresh
    def last_updated_by(self) -> Observer:
        return Observer(id_=int(self._raw_data["last_updated_by"]))

    @property  # type: ignore
    @check_refresh
    def last_updated_date(self) -> datetime:
        last_updated_date = datetime.fromtimestamp(
            int(self._raw_data["last_updated_date"]["@timestamp"])
            if type(self._raw_data["last_updated_date"]) is dict
            else int(self._raw_data["last_updated_date"]),
        ).astimezone()
        return last_updated_date

    # Following Properties appear only when requesting an Observation
    @property
    def municipality(self) -> str:
        return (
            self._raw_data["municipality"]
            if "municipality" in self._raw_data
            else self.local_admin_unit.name
        )

    @property  # type: ignore
    @check_refresh
    def county(self) -> Optional[str]:
        return self._raw_data["county"] if "county" in self._raw_data else None

    @property  # type: ignore
    @check_refresh
    def country(self) -> Optional[str]:
        return self._raw_data["country"] if "country" in self._raw_data else None

    # Following Properties appear only when requesting a Site
    @property
    def centroid(self) -> Optional[str]:
        return self._centroid

    @property
    def order(self) -> Optional[int]:
        return self._order

    @property
    def wkt(self) -> Optional[str]:
        return self._wkt

    @classmethod
    def find_closest_place(
        cls, coord_lat: float, coord_lon: float, get_hidden: bool = False, **kwargs
    ) -> "Place":
        return cls.list_all(
            find_closest_place="1",
            coord_lat=coord_lat,
            coord_lon=coord_lon,
            get_hidden=get_hidden,
            **kwargs,
        )[0]

    @classmethod
    def create_from_site(cls, data: Dict[str, Any]):
        identifier: int = int(data["id"])
        obj = cls(identifier)
        obj._centroid = data["centroid"] if "centroid" in data else None
        obj._order = data["order"] if "order" in data else None
        obj._wkt = data["wkt"] if "wkt" in data else None
        return obj

    @classmethod
    def diff(
        cls,
        date: datetime,
        modification_type: ModificationType = None,
        only_protocol: Union[str, BaseModel] = None,
        retrieve_places: bool = False,
        retries: int = 0,
    ) -> List["Place"]:
        """Retrieves a list of places which changed in between now and a given date
        :param date: Date in the past, to which changed places should be searched
        :param modification_type: Type of modification.
        :param only_protocol: Return only observation which are part of the given Protocol (Protocol Instance or Name)
        :param retrieve_places: Indicates if the place objects should be retrieved. Default: False
        :param retries: Indicates how many retries should be performed
        :type date: datetime
        :type modification_type: ModificationType
        :type only_protocol: Union[str, "Protocol"]
        :type retrieve_places: bool
        :type retries: int
        :return: List of observations
        :rtype: List[Observation]
        """
        url = f"{cls.ENDPOINT}/diff"
        params = dict()
        if modification_type:
            params["modification_type"] = modification_type.value
        if only_protocol:
            from ornitho.model.protocol import Protocol

            if isinstance(only_protocol, Protocol):
                params["only_protocol"] = only_protocol.name
            else:
                params["only_protocol"] = only_protocol.__str__()

        date = date.replace(microsecond=0)
        if date.tzinfo:
            date = date.astimezone(datetime.now().astimezone().tzinfo).replace(
                tzinfo=None
            )
        params[
            "date"
        ] = (
            date.isoformat()
        )  # Format here, because isoformat is mostly ignored, except here

        changed_places = cls.request(
            method="get", url=url, params=params, retries=retries
        )
        places = []
        for place in changed_places:
            if place["modification_type"] == "updated":
                modification_type = ModificationType.ONLY_MODIFIED
            else:
                modification_type = ModificationType.ONLY_DELETED
            if retrieve_places and modification_type == ModificationType.ONLY_MODIFIED:
                places.append(cls.get(int(place["id_place"])))
            else:
                places.append(
                    cls(id_=int(place["id_place"]), modification_type=modification_type)
                )
        return places



from datetime import datetime
from typing import List, Optional

from ornitho.model.abstract import ListableModel
from ornitho.model.abstract.base_model import check_refresh
from ornitho.model.right import Right


class Observer(ListableModel):
    ENDPOINT: str = "observers"

    def __init__(self, id_) -> None:
        """Observer constructor
        :param id_: ID, which is used to get the observer from Biolovison – None if a new observation will be created
        :type id_: int
        """
        super().__init__(id_)
        self._rights: List[Right] = []

    @classmethod
    def current(cls) -> "Observer":
        url = f"{cls.ENDPOINT}/current"
        response = cls.request(method="GET", url=url)
        return cls.create_from_ornitho_json(response[0])

    @property  # type: ignore
    @check_refresh
    def external_id(self) -> int:
        return int(self._raw_data["external_id"])

    @property  # type: ignore
    @check_refresh
    def name(self) -> str:
        return self._raw_data["name"]

    @property  # type: ignore
    @check_refresh
    def surname(self) -> str:
        return self._raw_data["surname"]

    @property  # type: ignore
    @check_refresh
    def street(self) -> str:
        return self._raw_data["street"]

    @property  # type: ignore
    @check_refresh
    def number(self) -> str:
        return self._raw_data["number"]

    @property  # type: ignore
    @check_refresh
    def postcode(self) -> str:
        return self._raw_data["postcode"]

    @property  # type: ignore
    @check_refresh
    def municipality(self) -> str:
        return self._raw_data["municipality"]

    @property  # type: ignore
    @check_refresh
    def lat(self) -> float:
        return float(self._raw_data["lat"])

    @property  # type: ignore
    @check_refresh
    def lon(self) -> float:
        return float(self._raw_data["lon"])

    @property  # type: ignore
    @check_refresh
    def email(self) -> str:
        return self._raw_data["email"]

    @property  # type: ignore
    @check_refresh
    def private_phone(self) -> str:
        return self._raw_data["private_phone"]

    @property  # type: ignore
    @check_refresh
    def work_phone(self) -> str:
        return self._raw_data["work_phone"]

    @property  # type: ignore
    @check_refresh
    def mobile_phone(self) -> str:
        return self._raw_data["mobile_phone"]

    @property  # type: ignore
    @check_refresh
    def birth_year(self) -> int:
        return int(self._raw_data["birth_year"])

    @property  # type: ignore
    @check_refresh
    def atlas_list(self) -> int:
        return int(self._raw_data["atlas_list"])

    @property  # type: ignore
    @check_refresh
    def id_universal(self) -> int:
        return int(self._raw_data["id_universal"])

    @property  # type: ignore
    @check_refresh
    def display_order(self) -> str:
        return self._raw_data["display_order"]

    @property  # type: ignore
    @check_refresh
    def registration_date(self) -> datetime:
        registration_date = datetime.fromtimestamp(
            int(self._raw_data["registration_date"]["@timestamp"]),
        ).astimezone()
        return registration_date

    @property  # type: ignore
    @check_refresh
    def last_inserted_data(self) -> Optional[datetime]:
        last_inserted_data = (
            datetime.fromtimestamp(
                int(self._raw_data["last_inserted_data"]["@timestamp"]),
            ).astimezone()
            if "last_inserted_data" in self._raw_data
            else None
        )
        return last_inserted_data

    @property  # type: ignore
    @check_refresh
    def last_login(self) -> datetime:
        last_login = datetime.fromtimestamp(
            int(self._raw_data["last_login"]["@timestamp"]),
        ).astimezone()
        return last_login

    @property  # type: ignore
    @check_refresh
    def anonymous(self) -> bool:
        return False if self._raw_data.get("anonymous") == "0" else True

    @property  # type: ignore
    @check_refresh
    def hide_email(self) -> bool:
        return False if self._raw_data.get("hide_email") == "0" else True

    @property  # type: ignore
    @check_refresh
    def photo(self) -> str:
        return self._raw_data["photo"]

    @property  # type: ignore
    @check_refresh
    def species_order(self) -> str:
        return self._raw_data["species_order"]

    @property  # type: ignore
    @check_refresh
    def langu(self) -> str:
        return self._raw_data["langu"]

    @property  # type: ignore
    @check_refresh
    def item_per_page_obs(self) -> int:
        return int(self._raw_data["item_per_page_obs"])

    @property  # type: ignore
    @check_refresh
    def item_per_page_gallery(self) -> int:
        return int(self._raw_data["item_per_page_gallery"])

    @property  # type: ignore
    @check_refresh
    def archive_account(self) -> bool:
        return False if self._raw_data.get("archive_account") == "0" else True

    @property  # type: ignore
    @check_refresh
    def collectif(self) -> bool:
        return False if self._raw_data.get("collectif") == "0" else True

    @property  # type: ignore
    @check_refresh
    def use_latin_search(self) -> bool:
        return False if self._raw_data.get("use_latin_search") == "N" else True

    @property  # type: ignore
    @check_refresh
    def private_website(self) -> str:
        return self._raw_data["private_website"]

    @property  # type: ignore
    @check_refresh
    def presentation(self) -> str:
        return self._raw_data["presentation"]

    @property  # type: ignore
    @check_refresh
    def has_search_access(self) -> bool:
        return False if self._raw_data.get("has_search_access") == "0" else True

    @property  # type: ignore
    @check_refresh
    def default_hidden(self) -> bool:
        return False if self._raw_data.get("default_hidden") == "0" else True

    @property  # type: ignore
    @check_refresh
    def debug_file_upload(self) -> bool:
        return False if self._raw_data.get("debug_file_upload") == "0" else True

    @property  # type: ignore
    @check_refresh
    def mobile_use_form(self) -> bool:
        return False if self._raw_data.get("mobile_use_form") == "0" else True

    @property  # type: ignore
    @check_refresh
    def mobile_use_mortality(self) -> bool:
        return False if self._raw_data.get("mobile_use_mortality") == "0" else True

    @property  # type: ignore
    @check_refresh
    def show_precise(self) -> bool:
        return False if self._raw_data.get("show_precise") == "0" else True

    @property  # type: ignore
    @check_refresh
    def bypass_purchase(self) -> bool:
        return False if self._raw_data.get("bypass_purchase") == "0" else True

    @property  # type: ignore
    @check_refresh
    def mobile_use_protocol(self) -> bool:
        return False if self._raw_data.get("mobile_use_protocol") == "0" else True

    @property  # type: ignore
    @check_refresh
    def mobile_use_trace(self) -> bool:
        return False if self._raw_data.get("mobile_use_trace") == "0" else True

    @property  # type: ignore
    def rights(self) -> List[Right]:
        if self.id_:
            if not self._rights:
                self._rights = Right.retrieve_for_observer(self.id_)
        return self._rights
